from .fabric import (
    BasicFabricTrainer,
    FabricTrainingContext,
    callbacks,
    default_fabric,
    procedure,
    train_epoch,
)
from .fabric.callbacks import (
    FabricCallbackProtocol,
    ParseCSVLogfile,
    PrintTrainInfo,
    SaveCheckpoints,
    SaveHparams,
    find_callback,
)
from .lightning_modules import BasicLightningTrainer
from .metric import TorchMetricRecorder, clf_metrics_builder, reg_metrics_builder
from .utils import cuda_available, default_device, seed_everything, use_device

train_loop = procedure.train_loop

__all__ = [
    'cuda_available',
    'default_device',
    'seed_everything',
    'use_device',
    'BasicLightningTrainer',
    'TorchMetricRecorder',
    'clf_metrics_builder',
    'reg_metrics_builder',
    'BasicFabricTrainer',
    'default_fabric',
    'callbacks',
    'procedure',
    'train_epoch',
    'find_callback',
    'FabricCallbackProtocol',
    'FabricTrainingContext',
    'ParseCSVLogfile',
    'PrintTrainInfo',
    'SaveCheckpoints',
    'SaveHparams',
    'train_loop'
]



import torch
import torch.nn as nn

from zyplib.nn.modules import BaseModule


class VerySimpleCNN(BaseModule):
    def __init__(
        self,
        in_channels,
        hid_channels,
        fc_layers=(1,),
        kernel_size=3,
        num_layers=3,
        activation=nn.LeakyReLU,
    ):
        super(VerySimpleCNN, self).__init__()

        self.layers = nn.ModuleList()
        for i in range(num_layers):
            self.layers.append(
                nn.Conv1d(
                    in_channels if i == 0 else hid_channels,
                    hid_channels,
                    kernel_size,
                    padding=kernel_size // 2,
                )
            )
            self.layers.append(activation())

        in_features = hid_channels
        fc = nn.ModuleList()
        for i in fc_layers:
            fc.append(nn.Linear(in_features, i))
            fc.append(activation())
            in_features = i
        fc = fc[:-1]  # 去掉最后一个激活函数
        self.fc = nn.Sequential(*fc)

    def forward(self, x):
        # x shape: (N, C, T)
        for layer in self.layers:
            x = layer(x)
        # x shape: (N, out_channels, T)
        x = x.mean(dim=2)  # Global average pooling
        # x shape: (N, out_channels)
        x = self.fc(x)
        # x = torch.sigmoid(x)
        return x


class AttentionCNN(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        att_channels,
        n_len_seg,
        n_classes
    ):
        """一个简单的自注意力机制的卷积神经网络

        @source: https://github.com/hsd1503/resnet1d/blob/master/acnn1d.py

        Input/Output
        ----------
        - Input: (batch, channels, length)
        - Output: (batch)

        Parameters
        ----------
        - `in_channels` : `int`
            - number of input channels
        - `out_channels` : `int`
            - number of output channels
        - `att_channels` : `int`
            - number of attention channels
        - `n_len_seg` : `int`
            - number of segments
        - `n_classes` : `int`
            - number of classes
        """
        super(AttentionCNN, self).__init__()

        self.n_len_seg = n_len_seg
        self.n_classes = n_classes
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.att_channels = att_channels

        # (batch, channels, length)
        self.cnn = nn.Conv1d(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=16,
            stride=4,
        )

        self.W_att_channel = nn.Parameter(
            torch.randn(self.out_channels, self.att_channels)
        )
        self.v_att_channel = nn.Parameter(torch.randn(self.att_channels, 1))

        self.dense = nn.Linear(out_channels, n_classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        self.n_channel, self.n_length = x.shape[-2], x.shape[-1]
        assert (
            self.n_length % self.n_len_seg == 0
        ), 'Input n_length should divided by n_len_seg'
        self.n_seg = self.n_length // self.n_len_seg

        out = x

        # (n_samples, n_channel, n_length) -> (n_samples, n_length, n_channel)
        out = out.permute(0, 2, 1)
        # (n_samples, n_length, n_channel) -> (n_samples*n_seg, n_len_seg, n_channel)
        out = out.view(-1, self.n_len_seg, self.n_channel)
        # (n_samples*n_seg, n_len_seg, n_channel) -> (n_samples*n_seg, n_channel, n_len_seg)
        out = out.permute(0, 2, 1)
        # cnn
        out = self.cnn(out)
        # global avg, (n_samples*n_seg, out_channels)
        out = out.mean(-1)
        # global avg, (n_samples, n_seg, out_channels)
        out = out.view(-1, self.n_seg, self.out_channels)
        # self attention
        e = torch.matmul(out, self.W_att_channel)
        e = torch.matmul(torch.tanh(e), self.v_att_channel)
        n1 = torch.exp(e)
        n2 = torch.sum(torch.exp(e), 1, keepdim=True)
        gama = torch.div(n1, n2)
        out = torch.sum(torch.mul(gama, out), 1)
        # dense
        out = self.dense(out)

        return out



import torch
import torch.nn as nn

from zyplib.nn.modules.base import BaseModule


class ChannelAttention(BaseModule):
    def __init__(self, in_ch, ratio=16):
        super(ChannelAttention, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool1d(1)
        self.max_pool = nn.AdaptiveMaxPool1d(1)

        self.fc = nn.Sequential(
            nn.Conv1d(in_ch, in_ch // ratio, 1, bias=False),
            nn.ReLU(),
            nn.Conv1d(in_ch // ratio, in_ch, 1, bias=False),
        )
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avg_out = self.fc(self.avg_pool(x))
        max_out = self.fc(self.max_pool(x))
        out = avg_out + max_out
        return self.sigmoid(out)


class SpatialAttention(BaseModule):
    def __init__(self, kernel_size=7):
        super(SpatialAttention, self).__init__()

        self.conv1 = nn.Conv1d(2, 1, kernel_size, padding=kernel_size // 2, bias=False)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        avg_out = torch.mean(x, dim=1, keepdim=True)
        max_out, _ = torch.max(x, dim=1, keepdim=True)
        x = torch.cat([avg_out, max_out], dim=1)
        x = self.conv1(x)
        return self.sigmoid(x)


class CBAM(BaseModule):
    def __init__(self, in_ch, ratio=16, kernel_size=7):
        super(CBAM, self).__init__()
        self.ca = ChannelAttention(in_ch, ratio)
        self.sa = SpatialAttention(kernel_size)

    def forward(self, x):
        x = x * self.ca(x)
        x = x * self.sa(x)
        return x



from typing import List, Tuple

import torch
from torch import nn


def output_shape(model: nn.Module, input_shape: Tuple[int, ...]) -> Tuple[int]:
    """计算模型输出形状

    Parameters
    ----------
    - `model` : `nn.Module`
        - 模型
    - `input_shape` : `Tuple[int, ...]`
        - 输入形状

    Returns
    ----------
    - `Tuple[int, ...]`
        - 输出形状
    """
    train_state = model.training
    shape = None
    with torch.no_grad():
        model.eval()
        shape = model(torch.randn(input_shape)).shape
    model.train(train_state)
    return shape



import numpy as np
import torch
from scipy import linalg


def calculate_frechet_distance(mu1, sigma1, mu2, sigma2, eps=1e-6):
    mu1 = np.atleast_1d(mu1)
    mu2 = np.atleast_1d(mu2)

    sigma1 = np.atleast_2d(sigma1)
    sigma2 = np.atleast_2d(sigma2)

    assert mu1.shape == mu2.shape, 'Training and test mean vectors have different lengths'
    assert (
        sigma1.shape == sigma2.shape
    ), 'Training and test covariances have different dimensions'

    diff = mu1 - mu2

    # Product might be almost singular
    covmean, _ = linalg.sqrtm(sigma1.dot(sigma2), disp=False)
    if not np.isfinite(covmean).all():
        msg = (
            'fid calculation produces singular product; '
            'adding %s to diagonal of cov estimates'
        ) % eps
        print(msg)
        offset = np.eye(sigma1.shape[0]) * eps
        covmean = linalg.sqrtm((sigma1 + offset).dot(sigma2 + offset))

    # Numerical error might give slight imaginary component
    if np.iscomplexobj(covmean):
        if not np.allclose(np.diagonal(covmean).imag, 0, atol=1e-2):
            m = np.max(np.abs(covmean.imag))
            raise ValueError('Imaginary component {}'.format(m))
        covmean = covmean.real

    tr_covmean = np.trace(covmean)

    return diff.dot(diff) + np.trace(sigma1) + np.trace(sigma2) - 2 * tr_covmean



import functools
import os
from dataclasses import dataclass
from typing import Any, Callable, Optional, TypeVar

import numpy as np
import torch
from diskcache import Cache, Timeout
from prettytable import PrettyTable

# from prettytable import PrettyTable
from zyplib._config import config
from zyplib.utils.print import print_info, print_warn, type_summary

_configs = {
    'cache_dir': config.DISK_CACHE_DIR,
    'cache_max_size': config.DISK_CACHE_MAX_SIZE, # MB
}


def update_cache_config(cache_dir: str = None, cache_max_size_mb: int | float = None):
    if cache_dir is not None:
        _configs['cache_dir'] = cache_dir
    if cache_max_size_mb is not None:
        _configs['cache_max_size'] = cache_max_size_mb


def cache_factory(cache_dir: str = None) -> Cache:
    """创建缓存对象

    - 如果传入 `cache_dir` 则使用传入的目录作为 diskcache 的缓存目录
    - 如果传入 `cache_dir` 为 None 则使用默认缓存目录
        - 默认缓存目录为 `./cache`, 可以通过 `update_cache_config` 函数设置

    Parameters
    ----------
    - `cache_dir` : `str`, optional
        - 缓存目录; 如果为 None 则使用 _config 中的默认缓存目录

    Returns
    ----------
    - `Cache`
        - 缓存对象
    """
    cache_dir = os.path.join(_configs['cache_dir']) if cache_dir is None else cache_dir
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)
    size_limit = _configs['cache_max_size'] * 1024 * 1024
    return Cache(cache_dir, size_limit=size_limit)


def get_cache(key: str, default: Any = None, cache_dir: str | None = None):
    """获取缓存

    Parameters
    ----------
    - `key` : `str`
        - 缓存键
    - `default` : `Any`, optional
        - 默认值，当缓存不存在时返回
    - `cache_dir` : `str | None`, optional
        - 缓存目录; 如果为 None 则使用 cache_factory 中的默认缓存目录

    Returns
    ----------
    - `Any`
        - 缓存值, 当缓存不存在或者过期时返回默认值
    """
    try:
        with cache_factory(cache_dir) as cache:
            return cache.get(key, default)
    except Timeout as t:
        print_warn(f'Cache[{key}] timeout: {t}')
        return default


def set_cache(
    key: str,
    value: Any,
    expire_seconds: int = None,
    tag: str = None,
    cache_dir: str | None = None,
):
    """设置缓存

    Parameters
    ----------
    - `key` : `str`
        - 缓存键
    - `value` : `Any`
        - 缓存值
    - `expire_seconds` : `int`, optional
        - 缓存过期时间; 如果为 None 则不设置过期时间
    - `tag` : `str`, optional
        - 缓存标签
    - `cache_dir` : `str | None`, optional
        - 缓存目录; 如果为 None 则使用 cache_factory 中的默认缓存目录

    Returns
    ----------
    - `bool`
        - 设置成功返回 True, 否则返回 False
    """
    with cache_factory(cache_dir) as cache:
        flag = cache.set(key, value, expire=expire_seconds, tag=tag)
    if not flag:
        print_warn(f'Cache[{key}] set failed')
    return flag


def del_cache(key: str, cache_dir: str | None = None):
    """删除缓存

    Parameters
    ----------
    - `key` : `str`
        - 缓存键
    - `cache_dir` : `str | None`, optional
        - 缓存目录; 如果为 None 则使用 cache_factory 中的默认缓存目录
    """
    with cache_factory(cache_dir) as cache:
        cache.delete(key)


def purge_all_cache(cache_dir: str | None = None):
    """清空所有缓存

    Parameters
    ----------
    - `cache_dir` : `str | None`, optional
        - 缓存目录; 如果为 None 则使用 cache_factory 中的默认缓存目录
    """
    with cache_factory(cache_dir) as cache:
        cache.clear()


def use_cache(
    key,
    fn_value: Callable[[], Any],
    cache_dir: str | None = None,
    expire_seconds: int = None,
    tag: str = None,
):
    """使用缓存, 如果缓存命中则直接返回缓存值, 否则调用 fn_value 计算值并缓存


    Parameters
    ----------
    - `key` : `str`
        - 缓存键
    - `fn_value` : `Callable[[], Any]`
        - 计算值的函数
    - `cache_dir` : `str | None`, optional
        - 缓存目录; 如果为 None 则使用 cache_factory 中的默认缓存目录
    - `expire_seconds` : `int`, optional
        - 缓存过期时间; 如果为 None 则不设置过期时间
    - `tag` : `str`, optional
        - 缓存标签

    Returns
    ----------
    - `Any`
        - 缓存值

    Examples
    ----------
    >>> def fn_value():
    >>>     return 1
    >>> use_cache('key', fn_value)
    """
    with cache_factory(cache_dir) as cache:
        value = cache.get(key, default=None)
        if value is None:
            value = fn_value()
            cache.set(key, value, expire=expire_seconds, tag=tag)
    return value


@dataclass
class CacheItem:
    key: str
    value: Any
    expire_time: int
    tag: str


def list_caches(cache_dir: str | None = None):
    """列出所有缓存; 通过

    Parameters
    ----------
    - `cache_dir` : `str`, optional
        - 缓存目录; 如果为 None 则使用 cache_factory 中的默认缓存目录

    Yields
    ----------
    - `CacheItem`: 缓存项
    """
    with cache_factory(cache_dir) as cache:
        keys = list(cache.iterkeys())
        for key in keys:
            obj, expire_time, tag = cache.get(
                key, default=None, expire_time=True, tag=True
            )
            cache_item = CacheItem(key=key, value=obj, expire_time=expire_time, tag=tag)
            yield cache_item


def display_caches(cache_dir: str | None = None, table: bool = True):
    """打印缓存状态

    Parameters
    ----------
    - `cache_dir` : `str`, optional
        - 缓存目录; 如果为 None 则使用 cache_factory 中的默认缓存目录
    - `table` : `bool`, optional
        - 是否使用 PrettyTable 展示; 如果为 False 则使用 print 打印
    """
    with cache_factory(cache_dir) as cache:
        size = cache.volume()
    print_info(f'缓存大小: {size / 1024 / 1024:.2f} MB')

    caches = list_caches(cache_dir)
    if table:
        table = PrettyTable(['Index', 'Key', 'Value', 'Expire', 'Tag'])
        for idx, cache_item in enumerate(caches):
            table.add_row(
                [
                    idx,
                    type_summary(cache_item.key),
                    type_summary(cache_item.value),
                    cache_item.expire_time,
                    cache_item.tag,
                ]
            )
        print_info(f'缓存: {_configs["cache_dir"]}'.center(60, '='))
        print_info(table)
    else:
        print_info(f'缓存: {_configs["cache_dir"]}')
        for idx, cache_item in enumerate(caches, start=1):
            print_info(f'缓存 {idx}'.center(60, '='))
            print_info('Key:', end='\t')
            print(cache_item.key)
            print_info('Value:', end='\t')
            print(cache_item.value)
            print_info('Expire:', end='\t')
            print(cache_item.expire_time)
            print_info('Tag:', end='\t')
            print(cache_item.tag)
            print('')


F = TypeVar('F', bound=Callable)


def make_func_memoize(
    func: F, cache: Cache, name=None, typed=False, expire=None, tag=None, ignore=()
) -> F:
    """接受一个函数对象，为其动态添加 diskcache 的 memoize 功能"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        memoized_func = cache.memoize(
            name=name, typed=typed, expire=expire, tag=tag, ignore=ignore
        )(func)
        return memoized_func(*args, **kwargs)

    return wrapper



from copy import deepcopy
from typing import Any, Callable, Dict, List, Literal, Optional, Union

import torch
from lightning.fabric import Fabric
from torch import nn
from torch.optim.lr_scheduler import _LRScheduler
from torch.utils.data import DataLoader
from torchmetrics import Metric, MetricCollection

from zyplib.train.fabric.callbacks import FabricLogMetrics
from zyplib.train.metric.torch_metric import TorchMetricRecorder
from zyplib.train.utils import prefixed_dict, step_lr_sched, use_device
from zyplib.utils.print import print_info
from zyplib.utils.progress_bar import rich_progress


class BasicFabricTrainer:
    def __init__(
        self,
        fabric: Fabric,
        model: nn.Module,
        optimizer: torch.optim.Optimizer,
        fn_loss: nn.Module = None,
        hparams: dict[str, Any] | None = None,
        metric_builder: Callable[[], MetricCollection | dict[str, Metric]] | None = None,
        before_metric: Literal['sigmoid', 'softmax']
        | Callable[[torch.Tensor], torch.Tensor]
        | dict[str | Literal['*'], Callable] = None,
        lr_scheduler: Optional[_LRScheduler] = None,
        lr_sched_freq: int = 1,
        lr_sched_interval: str = 'epoch',
    ):
        """基于 Lightning Fabric 的基础训练器

        提供了基本的训练、验证功能，支持评估指标记录和学习率调度等功能

        Parameters
        ----------
        - `fabric` : `Fabric`
            - Lightning Fabric 实例
            - 用于管理分布式训练、混合精度等
        - `model` : `nn.Module`
            - PyTorch 模型
        - `optimizer` : `torch.optim.Optimizer`
            - 优化器实例
        - `fn_loss` : `nn.Module`, optional
            - 损失函数
            - 为 None 时需要在训练步骤中手动计算损失
        - `hparams` : `dict[str, Any] | None`, optional
            - 超参数字典
            - 会被深拷贝以避免修改原始值
        - `metric_builder` : `Callable[[], MetricCollection | dict[str, Metric]] | None`, optional
            - 评估指标构建函数, 每次调用返回新的评估指标实例，用于训练集和验证集
            - 返回 MetricCollection 或 dict[str, Metric] 类型
            - 会自动移动到 fabric 所在的设备上, 无需手动 to('cuda')
        - `before_metric` : `'sigmoid' | 'softmax' | Callable | dict[str | '*', Callable]`, optional
            - 在计算评估指标之前的激活函数: `(y_hat) -> y_activated`
            - 支持预定义的 'sigmoid' 和 'softmax'
            - 支持自定义激活函数
            - 支持为不同指标指定不同的激活函数
        - `lr_scheduler` : `Optional[_LRScheduler]`, optional
            - 学习率调度器
        - `lr_sched_freq` : `int`, optional
            - 学习率调度的频率
            - 每隔多少个间隔更新一次学习率
        - `lr_sched_interval` : `str`, optional
            - 学习率调度的间隔单位
            - 可选值: 'epoch' 或 'step'

        Example
        ----------
        >>> def build_metrics():
        ...     return MetricCollection({
        ...         'accuracy': Accuracy(task='binary'),
        ...         'f1': F1Score(task='binary')
        ...     })
        >>>
        >>> trainer = BasicFabricTrainer(
        ...     fabric=fabric,
        ...     model=model,
        ...     optimizer=optimizer,
        ...     fn_loss=nn.BCEWithLogitsLoss(),
        ...     metric_builder=build_metrics,
        ...     before_metric='sigmoid',
        ...     lr_scheduler=lr_scheduler,
        ...     lr_sched_interval='epoch'
        ... )
        """
        self.fabric = fabric
        self.fabric.launch()

        self.hparams = deepcopy(hparams) if hparams is not None else None

        # 不使用同一个 Metric 对象, 避免麻烦问题
        metrics = metric_builder() if metric_builder is not None else None
        for metric in metrics.values():
            use_device(metric, self.fabric.device)

        self.train_metric = TorchMetricRecorder(metrics, before_metric=before_metric)

        metrics = metric_builder() if metric_builder is not None else None
        for metric in metrics.values():
            use_device(metric, self.fabric.device)

        self.val_metric = TorchMetricRecorder(metrics, before_metric=before_metric)

        self.model = model
        self.optimizer = optimizer
        self.fn_loss = fn_loss or nn.BCEWithLogitsLoss()
        self.lr_scheduler = lr_scheduler
        self.lr_sched_freq = lr_sched_freq
        self.lr_sched_interval = lr_sched_interval

        self.fabric_log_metrics = FabricLogMetrics(fabric)
        self.fabric._callbacks.append(self.fabric_log_metrics)

        self.early_stop = False  # 中断训练, 给 EarlyStopping 使用

    def training_step(
        self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> tuple[torch.Tensor, float, Dict[str, Any]]:
        self.model.train()
        x, y = batch
        self.optimizer.zero_grad()
        y_pred = self.model(x)
        loss = self.fn_loss(y_pred, y)
        self.fabric.backward(loss)
        self.optimizer.step()

        metrics = self.train_metric.update(y_pred, y, loss.item())
        return y_pred, loss, metrics

    def validation_step(
        self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> tuple[torch.Tensor, float, Dict[str, Any]]:
        self.model.eval()
        with torch.no_grad():
            x, y = batch
            y_pred = self.model(x)
            loss = self.fn_loss(y_pred, y)

        metrics = self.val_metric.update(y_pred, y, loss.item())
        return y_pred, loss, metrics

    def test_step(
        self, batch: tuple[torch.Tensor, torch.Tensor], batch_idx: int
    ) -> tuple[torch.Tensor, float, Dict[str, Any]]:
        self.model.eval()
        with torch.no_grad():
            x, y = batch
            y_pred = self.model(x)
            loss = self.fn_loss(y_pred, y)

        metrics = self.val_metric.update(y_pred, y, loss.item())
        return y_pred, loss, metrics

    def train_epoch(self, train_loader: DataLoader, epoch_idx: int) -> Dict[str, float]:
        for batch_idx, batch in enumerate(
            rich_progress(train_loader, 'Training', color='green')
        ):
            y_pred, loss, metrics = self.training_step(batch, batch_idx)
            # 保持与原代码相同的callback调用时机
            self._callbacks(
                'on_train_batch_end',
                metrics=metrics,
                y_pred=y_pred,
                batch=batch,
                loss=loss.item(),
                batch_idx=batch_idx,
            )
        metrics = self.train_metric.compute()
        return metrics

    def val_epoch(self, val_loader: DataLoader, epoch_idx: int) -> Dict[str, float]:
        for batch_idx, batch in enumerate(
            rich_progress(val_loader, 'Validating', color='blue')
        ):
            y_pred, loss, metrics = self.validation_step(batch, batch_idx)
            # 保持与原代码相同的callback调用时机
            self._callbacks(
                'on_validation_batch_end',
                metrics=metrics,
                y_pred=y_pred,
                batch=batch,
                loss=loss.item(),
                batch_idx=batch_idx,
            )
        metrics = self.val_metric.compute()
        return metrics

    def test_epoch(self, test_loader: DataLoader, epoch_idx: int) -> Dict[str, float]:
        for batch_idx, batch in enumerate(
            rich_progress(test_loader, 'Testing', color='blue')
        ):
            y_pred, loss, metrics = self.test_step(batch, batch_idx)
            # 保持与原代码相同的callback调用时机
            self._callbacks(
                'on_test_batch_end',
                metrics=metrics,
                y_pred=y_pred,
                batch=batch,
                loss=loss.item(),
                batch_idx=batch_idx,
            )
        metrics = self.val_metric.compute()
        return metrics

    def fit(
        self,
        train_loader: DataLoader,
        val_loader: Optional[DataLoader] = None,
        max_epochs: int = 10,
    ) -> Dict[str, List[float]]:
        self._callbacks('on_fit_start', max_epochs=max_epochs)

        self.model, self.optimizer = self.fabric.setup(self.model, self.optimizer)
        train_loader = self.fabric.setup_dataloaders(train_loader)
        if val_loader:
            val_loader = self.fabric.setup_dataloaders(val_loader)

        self._callbacks(
            'on_train_start',
            max_epochs=max_epochs,
            hparams=self.hparams,
            model=self.model,
            optimizer=self.optimizer,
            lr_scheduler=self.lr_scheduler,
            fabric=self.fabric,
        )

        for epoch in range(max_epochs):
            if self.early_stop:
                print_info('停止训练')
                break

            self._callbacks('on_train_epoch_start', epoch=epoch, max_epochs=max_epochs)

            train_metrics = self.train_epoch(train_loader, epoch)

            self._callbacks('on_train_epoch_end', metrics=train_metrics, epoch=epoch)

            if val_loader:
                self._callbacks('on_validation_epoch_start', epoch=epoch)

                val_metrics = self.val_epoch(val_loader, epoch)

                self._callbacks(
                    'on_validation_epoch_end', metrics=val_metrics, epoch=epoch
                )
            else:
                val_metrics = None

            if self.lr_scheduler and self.lr_sched_interval == 'epoch':
                if (epoch + 1) % self.lr_sched_freq == 0:
                    loss = train_metrics['loss']
                    if val_metrics:
                        loss = val_metrics['loss']
                    step_lr_sched(self.lr_scheduler, loss)

            self._callbacks(
                'on_epoch_end',
                epoch=epoch,
                train_metrics=train_metrics,
                val_metrics=val_metrics,
            )

        self._callbacks('on_train_end')
        self._callbacks('on_fit_end')

    def test(self, test_loader: DataLoader) -> Dict[str, float]:
        self._callbacks('on_test_start', dataloader=test_loader)
        test_loader = self.fabric.setup_dataloaders(test_loader)
        result = self.test_epoch(test_loader, 0)
        self._callbacks('on_test_end', metrics=result)
        return result

    def _callbacks(self, hook_name: str, **kwargs):
        """调用 fabric 的 callbacks

        Parameters
        ----------
        - `hook_name` : `str`
            - 钩子名称
        - `kwargs` : `Dict[str, Any]`
            - 传入的 hook 参数
        """
        self.fabric.call(hook_name, trainer=self, **kwargs)



import os
import tkinter as tk
from pathlib import Path
from tkinter import filedialog, ttk
from typing import Callable

from ._utils import set_dpi_awareness


def launch_app(fn_file_cb: Callable[[Path], None]):
    def select_file():
        extensions = extension_entry.get().strip()
        filetypes = [('All files', '*.*')]
        if extensions:
            ext_list = [f'*.{ext.strip()}' for ext in extensions.split(',')]
            filetypes = [('Selected files', ' '.join(ext_list))] + filetypes

        file_path = filedialog.askopenfilename(
            initialdir=os.getcwd(), filetypes=filetypes
        )
        if file_path:
            root.destroy()
            fn_file_cb(Path(file_path))

    set_dpi_awareness()
    root = tk.Tk()
    root.title('选择文件')
    root.geometry('400x150')

    frame = ttk.Frame(root, padding='10')
    frame.pack(fill=tk.BOTH, expand=True)

    extension_label = ttk.Label(frame, text='文件后缀 (用逗号 "," 分隔):')
    extension_label.pack(pady=(0, 5))

    extension_entry = ttk.Entry(frame)
    extension_entry.pack(fill=tk.X, pady=(0, 10))
    extension_entry.insert(0, 'mat,edf')

    button = ttk.Button(frame, text='选择文件', command=select_file)
    button.pack(expand=True)

    root.mainloop()


if __name__ == '__main__':

    def fn_file_cb(path: Path):
        print(path)

    launch_app(fn_file_cb)



from .eegnet import EEGNet
from .lenet import LeNetPro
from .simple_cnn import AttentionCNN, VerySimpleCNN
from .simple_rnn import CRNN
from .tcn import TemporalConvNet

__all__ = [
    "EEGNet",
    "LeNetPro",
    "VerySimpleCNN",
    "CRNN",
    "TemporalConvNet",
    "AttentionCNN",
]



from datetime import datetime, timedelta
from typing import List, Optional, Set, TypedDict, Union
from xml.etree import ElementTree as ET

from zyplib.annotation.event import Event


def write_annotationlist_xml(events: List[Event], start_time: datetime, xml_fpath: str):
    """导出 EDFBrowser 的 xml 格式标注文件

    Parameters
    ----------
        - `events` : List[Event]
            - 事件列表, Evnet 为 zyplib.annotation.type.Event 类型; 单位为秒
        - `start_time` : datetime
            - 时间开始时间
        - `xml_fpath` : str
            - 导出的 xml 文件路径
    """
    annotationlist = ET.Element('annotationlist')
    for event in events:
        annotation = ET.SubElement(annotationlist, 'annotation')
        onset = event.onset
        duration = event.duration
        description = event.name
        onset_time = start_time + timedelta(seconds=onset)
        onset_time_str = onset_time.strftime('%Y-%m-%dT%H:%M:%S.%f')
        ET.SubElement(annotation, 'onset').text = onset_time_str
        ET.SubElement(annotation, 'duration').text = str(duration)
        ET.SubElement(annotation, 'description').text = description
    xml_bytes = ET.tostring(
        annotationlist, encoding='UTF-8', xml_declaration=True
    ).replace(b"'", b'"')
    with open(xml_fpath, 'wb') as f:
        f.write(xml_bytes)



from .filter import bandpass_butter, notch
from .resample import resample_eeg
from .scale import minmax, robust_scale, scale_eeg, z_score
from .segment import segment_signal

__all__ = [
    'segment_signal',
    'notch',
    'bandpass_butter',
    'scale_eeg',
    'resample_eeg',
    'z_score',
    'minmax',
    'robust_scale',
]



import numpy as np
import torch
from torch import fft
from torch.nn import functional as F

from zyplib.utils.ensure import ensure_3dims


def _periodogram(X: torch.Tensor, fs, detrend, scaling):
    if X.dim() > 2:
        X = torch.squeeze(X)
    elif X.dim() == 1:
        X = X.unsqueeze(0)

    if detrend:
        X -= X.mean(-1, keepdim=True)

    N = X.size(-1)
    assert N % 2 == 0

    df = fs / N
    dt = df
    f = torch.arange(0, N / 2 + 1) * df  # [0:df:f/2]

    dual_side = fft.fft(X)  # 双边谱
    half_idx = int(N / 2 + 1)
    single_side = dual_side[:, 0:half_idx]
    win = torch.abs(single_side)

    ps = win**2
    if scaling == 'density':  # 计算功率谱密度
        scale = N * fs
    elif scaling == 'spectrum':  # 计算功率谱
        scale = N**2
    elif scaling is None:  # 不做缩放
        scale = 1
    else:
        raise ValueError('Unknown scaling: %r' % scaling)
    Pxy = ps / scale

    Pxy[:, 1:-1] *= 2  # 能量2倍;直流分量不用二倍, 中心频率点不用二倍

    return f, Pxy.squeeze()


def periodogram(X: torch.Tensor, fs=256, detrend=True, scaling='density', no_grad=True):
    """计算信号单边 PSD, 基本等价于 scipy.signal.periodogram

    Parameters:
    ----------
        - `X`:          torch.Tensor, EEG, [T]/[N, T]
        - `fs`:         int, 采样率, Hz
        - `detrend`:    bool, 是否去趋势 (去除直流分量)
        - `scaling`:    { 'density', 'spectrum' }, 可选
            - 'density':    计算功率谱密度 `(V ** 2 / Hz)`
            - 'spectrum':    计算功率谱 `(V ** 2)`
        - `no_grad`:    bool, 是否启用 no_grad() 模式

    Returns:
    ----------
        - `Pxy`:    Tensor, 单边功率谱
    """
    if no_grad:
        with torch.no_grad():
            return _periodogram(X, fs, detrend, scaling)
    else:
        return _periodogram(X, fs, detrend, scaling)


def _get_window(window, nwlen, device):
    if window == 'hann':
        window = torch.hann_window(
            nwlen, dtype=torch.float32, device=device, periodic=False
        )
    elif window == 'hamming':
        window = torch.hamming_window(
            nwlen, dtype=torch.float32, device=device, periodic=False
        )
    elif window == 'blackman':
        window = torch.blackman_window(
            nwlen, dtype=torch.float32, device=device, periodic=False
        )
    elif window == 'boxcar':
        window = torch.ones(nwlen, dtype=torch.float32, device=device)
    else:
        raise ValueError('Invalid Window {}' % window)
    return window


def _pwelch(X: torch.Tensor, fs, detrend, scaling, window, nwlen, nhop):
    X = ensure_3dims(X, newaxis='channel')
    if scaling == 'density':
        scale = fs * (window * window).sum().item()
    elif scaling == 'spectrum':
        scale = window.sum().item() ** 2
    else:
        raise ValueError('Unknown scaling: %r' % scaling)
    # --------------- Fold and windowing --------------- #
    N, T = X.size(0), X.size(-1)
    X = X.view(N, 1, 1, T)
    X_fold = F.unfold(X, (1, nwlen), stride=nhop)  # [N, 1, 1, T] -> [N, nwlen, win_cnt]
    if detrend:
        X_fold -= X_fold.mean(1, keepdim=True)  # 各个窗口各自detrend
    window = window.view(1, -1, 1)  # [1, nwlen, 1]
    X_windowed = X_fold * window  # [N, nwlen, win_cnt]
    win_cnt = X_windowed.size(-1)

    # --------------- Pwelch --------------- #
    X_windowed = X_windowed.transpose(1, 2).contiguous()  # [N, win_cnt, nwlen]
    X_windowed = X_windowed.view(N * win_cnt, nwlen)  # [N * win_cnt, nwlen]
    f, pxx = _periodogram(
        X_windowed, fs, detrend=False, scaling=None
    )  # [N * win_cnt, nwlen // 2 + 1]
    pxx /= scale
    pxx = pxx.view(N, win_cnt, -1)  # [N, win_cnt, nwlen // 2 + 1]
    pxx = torch.mean(pxx, dim=1)  # [N, nwlen // 2 + 1]
    return f, pxx


def pwelch(
    X: torch.Tensor,
    fs=256,
    detrend=True,
    scaling='density',
    window='hann',
    nwlen=128,
    nhop=None,
    noverlap=None,
    no_grad=True,
):
    """Pwelch 方法，大致相当于 scipy.signal.welch

    Parameters:
    ----------
        - `X`:          torch.Tensor, EEG, [T]/[N, T]
        - `fs`:         int, 采样率, Hz
        - `detrend`:    bool, 是否去趋势 (去除直流分量)
        - `scaling`:    { 'density', 'spectrum' }, 可选
            - 'density':    计算功率谱密度 `(V ** 2 / Hz)`
            - 'spectrum':    计算功率谱 `(V ** 2)`
        - `window`:     str, 窗函数名称
        - `nwlen`:      int, 窗函数长度 (点的个数)
        - `nhop`:       int, 窗函数移动步长, 即 nwlen - noverlap (点的个数)
                        如果为 None 且 noverlap 为 None，则默认为 `nwlen // 2`
        - `noverlap`:   int, 仅当 nhop 为 None 时有效
        - `no_grad`:    bool, 是否启用 no_grad() 模式

    Returns:
    ----------
        - `f'`:    Tensor, 频率
        - `Pxy`:    Tensor, 单边功率谱
    """
    if nhop is None:
        if noverlap is not None:
            nhop = nwlen - noverlap
        else:
            nhop = nwlen // 2
    window = _get_window(window, nwlen, X.device)
    if no_grad:
        with torch.no_grad():
            return _pwelch(X, fs, detrend, scaling, window, nwlen, nhop)
    else:
        return _pwelch(X, fs, detrend, scaling, window, nwlen, nhop)


def _stft(X: torch.Tensor, fs, window, nhop, nfft):
    nfft = window.size(-1)
    # [batch, freq_points, time_points]
    stft_ans = torch.stft(
        X, n_fft=nfft, window=window, hop_length=nhop, return_complex=True
    )
    pxx = torch.abs(stft_ans) ** 2
    f = torch.arange(0, stft_ans.size(-2))
    ts = nhop / fs
    t = torch.arange(0, stft_ans.size(-1)) * ts
    return t, f, pxx


def stft(
    X: torch.Tensor, fs, window='hann', nwlen=256, nhop=None, nfft=512, no_grad=True
):
    """Short Time Fourier Transform

    Parameters:
    ----------
        - `X`:          torch.Tensor, EEG, [T]/[N, T]
        - `fs`:         int, 采样率, Hz
        - `window`:     str, 窗函数名称
        - `nhop`:       int, 窗函数移动步长, 即 nwlen - noverlap (点的个数)
                        如果为 None 且 noverlap 为 None，则默认为 `nwlen // 2`
        - `nfft`:       int, FFT 长度
        - `no_grad`:    bool, 是否启用 no_grad() 模式

    Returns:
    ----------
        - `pxx`: Tensor, 功率谱
    """
    if nhop is None:
        nhop = nwlen // 4
    if nfft < nwlen:
        nfft = nwlen
    window = _get_window(window, nwlen, X.device)
    if no_grad:
        with torch.no_grad():
            return _stft(X, fs, window, nhop, nfft)
    else:
        return _stft(X, fs, window, nhop, nfft)


def sinc(x: torch.Tensor):
    return torch.where(
        x == 0, torch.tensor(1.0, device=x.device, dtype=x.dtype), torch.sin(x) / (x)
    )


def gate_signal(cutoff, half_size=256):
    window = torch.hann_window(2 * half_size, periodic=False)
    time = torch.arange(-half_size, half_size)
    gate = 2 * cutoff * window * sinc(2 * cutoff * np.pi * time)
    gate /= gate.sum()
    return gate



from typing import Any, Callable, Dict, List, Literal, Optional, Union

import torch
from torch import nn
from torch.optim.lr_scheduler import _LRScheduler
from torch.utils.data import DataLoader
from torchmetrics import Metric, MetricCollection

from zyplib.train.metric.torch_metric import TorchMetricRecorder
from zyplib.train.utils import step_lr_sched, use_device
from zyplib.utils.print import print_info
from zyplib.utils.progress_bar import rich_progress


class BasicTrainer:
    def __init__(
        self,
        model: nn.Module,
        optimizer: torch.optim.Optimizer,
        fn_loss: nn.Module = None,
        metrics: Optional[Union[MetricCollection, Dict[str, Metric]]] = None,
        device: str = 'cuda' if torch.cuda.is_available() else 'cpu',
        lr_scheduler: Optional[_LRScheduler] = None,
        lr_sched_freq: int = 1,
        lr_sched_interval: str = 'epoch',
        before_metric: Optional[
            Union[Literal['sigmoid', 'softmax'], Callable, nn.Module]
        ] = None,
    ):
        self.train_metric = TorchMetricRecorder(metrics, before_metric=before_metric)
        self.val_metric = TorchMetricRecorder(metrics, before_metric=before_metric)

        self.model = use_device(model, device)
        self.optimizer = optimizer
        self.fn_loss = fn_loss or nn.BCEWithLogitsLoss()
        self.device = device
        self.lr_scheduler = lr_scheduler
        self.lr_sched_freq = lr_sched_freq
        self.lr_sched_interval = lr_sched_interval

    def training_step(self, batch: tuple[torch.Tensor, torch.Tensor]) -> Dict[str, Any]:
        self.model.train()
        x, y = use_device(batch[0], self.device), use_device(batch[1], self.device)

        self.optimizer.zero_grad()
        y_pred = self.model(x)
        loss = self.fn_loss(y_pred, y)
        loss.backward()
        self.optimizer.step()

        return self.train_metric.update(y_pred, y, loss.item())

    def validation_step(self, batch: Dict[str, torch.Tensor]) -> Dict[str, Any]:
        self.model.eval()
        with torch.no_grad():
            x, y = use_device(batch[0], self.device), use_device(batch[1], self.device)
            y_pred = self.model(x)
            loss = self.fn_loss(y_pred, y)

        return self.val_metric.update(y_pred, y, loss.item())

    def train_epoch(self, train_loader: DataLoader) -> Dict[str, float]:
        for batch in rich_progress(train_loader, 'Training', color='green'):
            self.training_step(batch)
        return self.train_metric.compute()

    def val_epoch(self, val_loader: DataLoader) -> Dict[str, float]:
        for batch in rich_progress(val_loader, 'Validating', color='blue'):
            self.validation_step(batch)
        return self.val_metric.compute()

    def fit(
        self,
        train_loader: DataLoader,
        val_loader: Optional[DataLoader] = None,
        num_epochs: int = 10,
    ) -> Dict[str, List[float]]:
        for epoch in range(num_epochs):
            print_info(f'Epoch {epoch+1}/{num_epochs}'.center(50, '='))
            train_metrics = self.train_epoch(train_loader)

            if val_loader:
                val_metrics = self.val_epoch(val_loader)
            else:
                val_metrics = None

            print_info('\tTrain: ', end='')
            self._print_epoch_summary(train_metrics)
            if val_loader:
                print_info('\tVal: ', end='')
                self._print_epoch_summary(val_metrics)
            print()

            if self.lr_scheduler and self.lr_sched_interval == 'epoch':
                if (epoch + 1) % self.lr_sched_freq == 0:
                    loss = train_metrics['loss']
                    if val_metrics:
                        loss = val_metrics['loss']
                    step_lr_sched(self.lr_scheduler, loss)

    def test(self, test_loader: DataLoader) -> Dict[str, float]:
        result = self.val_epoch(test_loader)
        print_info('\tTest: ', end='')
        self._print_epoch_summary(result)
        return result

    def _print_epoch_summary(self, metrics: Dict[str, float]):
        for key, value in metrics.items():
            print_info(f'\t{key}: {value:.4f}', end='')
        print()



import inspect
import os
import random
from functools import lru_cache
from typing import Optional, Union

import numpy as np
import torch
from simple_inject import inject
from torch import nn

from zyplib.utils.print import print_debug


def seed_everything(seed: int):
    os.environ['PL_GLOBAL_SEED'] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def step_lr_sched(
    scheduler: torch.optim.lr_scheduler.LRScheduler,
    metrics: Optional[Union[float, torch.Tensor]] = None,
):
    """step lr scheduler"""
    # scheduler.step()
    if isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
        scheduler.step(metrics)
    else:
        scheduler.step()


def prefixed_dict(prefix: str, d: dict):
    """给字典的每个键值对添加前缀

    >>> dic = torch_metric.compute()
    >>> prefixed_dict('val_', dic)
    >>> {'val_loss': 0.1, 'val_acc': 0.9}


    Parameters
    ----------
    - `prefix` : `str`
        - 前缀
    - `d` : `dict`
        - 需要添加前缀的字典

    Returns
    ----------
    - `dict`
        - 添加前缀后的字典
    """
    return {f'{prefix}{k}': v for k, v in d.items()}


@lru_cache
def cuda_available():
    return torch.cuda.is_available()


@lru_cache
def default_device():
    return torch.device('cuda' if cuda_available() else 'cpu')


def use_device(
    x: Union[torch.Tensor, nn.Module],
    device: Optional[Union[str, torch.device]] = None,
    hint: bool = False,
):
    """使用指定的设备

    依赖注入:
    ----------
    device 如果为 None，则首先会 inject('device') 来尝试获取设备名称，如果仍然为 None，则使用默认设备

    Parameters
    ----------
    - `x` : `torch.Tensor | nn.Module`
        - 需要移动的模型或张量
    - `device` : `str | torch.device`, optional
        - 设备名称或设备对象，如果为None，则使用默认设备
    - `hint` : `bool`, optional
        - 是否打印移动信息，默认不打印

    Returns
    ----------
    - `torch.Tensor | nn.Module`
        - 移动后的模型或张量
    """
    if not isinstance(x, (torch.Tensor, nn.Module)):
        return x
    elif not hasattr(x, 'to'):
        return x

    if device is None:
        device = inject('device', if_not_found='none')
    if device is None:
        device = default_device()

    if hint:
        print_debug(f'use_device: {x.__class__.__name__} --> {device}')
    return x.to(device)



import sys


def parse_cli_args():
    """解析任意命令行中 key=value 形式的参数

    只要是后面跟着的 --argname=argvalue 都会被解析到字典中
    """

    def _parse_str_to_value(value: str):
        try:
            return int(value)
        except ValueError:
            ...

        try:
            return float(value)
        except ValueError:
            ...

        if value.lower() in ('true', 'false'):
            return value.lower() == 'true'

        try:
            return eval(value)
        except Exception:
            ...

        return value

    args = {}
    for arg in sys.argv[1:]:
        if '=' in arg:
            key, value = arg.split('=')
            if key.startswith('--'):
                key = key[2:]
            args[key] = _parse_str_to_value(value)
    return args



from typing import Callable, Dict, List, Literal, Optional

# import matplotlib.pyplot as plt
import torch
from torchmetrics import Metric, MetricCollection

from zyplib.train.utils import use_device


class TorchMetricRecorder:
    def __init__(
        self,
        metrics: MetricCollection | dict[str, Metric],
        before_metric: Literal['sigmoid', 'softmax']
        | Callable[[torch.Tensor], torch.Tensor]
        | dict[str | Literal['*'], Callable] = None,
    ):
        """添加基于 `torchmetrics` 的日志评估指标的功能

        Parameters
        ----------
        - `metrics` : `MetricCollection | Dict[str, Metric]`, optional
            - 评估指标
            - 使用 `torchmetrics.MetricCollection` 或 `dict` 类型
        - `before_metric` : `'sigmoid' | 'softmax' | Callable | dict[str, Callable]`, optional

        `before_metric`:
        ----------
        在计算评估指标之前的激活函数: `(y_hat) -> y_activated`

        例如在二分类问题中，模型最后一层没有激活函数, 在计算评估指标 (即 `metrics`) 之前需要使用 `torch.sigmoid` 激活

        - `None`: 不需要额外的激活函数
        - `Literal['sigmoid', 'softmax']`: 使用 torch.sigmoid 和 torch.softmax
        - `(Tensor) -> Tensor`: 指定的激活函数
        - `dict[str, (Tensor) -> Tensor]`: 在比较复杂的情况下, 各个指标使用的输入不同的情况下可以用这个
            - key: 如果为 `*` 则代表通配符
            - value: `(Tensor) -> Tensor`, 为 None，则不激活
        """
        if before_metric == 'sigmoid':
            self.before_metric = torch.sigmoid
        elif before_metric == 'softmax':
            self.before_metric = torch.softmax
        else:
            self.before_metric = before_metric

        if metrics is None:
            self.metrics = MetricCollection({})
        elif isinstance(metrics, dict):
            self.metrics = MetricCollection(metrics)
        else:
            self.metrics = metrics

        self.step_history = {metric_name: [] for metric_name in self.metrics.keys()}
        self.step_history['loss'] = []
        self.epoch_history = []

        self.current_epoch = 0
        self.current_step = 0

        self.epoch_metrics = {
            metric_name: metric.clone() for metric_name, metric in self.metrics.items()
        }
        self.epoch_losses = []

    def to_device(self, device: str | torch.device):
        for metrics in [self.metrics, self.epoch_metrics]:
            for key, metric in metrics.items():
                # self.metrics[key] = use_device(metric, device)
                use_device(metric, device)

    def update(self, y_hat, y, loss: Optional[float] = None):
        """更新某个 step 的评估指标

        Parameters
        ----------
        - `y_hat` : `torch.Tensor`
            - 预测值
        - `y` : `torch.Tensor`
            - 真实值
        - `loss` : `float?`, optional
            - 损失值, 注意是 `float` 类型, 需要使用 `item()` 方法获取
        """
        result = {}

        # Log loss if provided
        if loss is not None:
            if isinstance(loss, torch.Tensor):
                loss = loss.item()
            self.step_history['loss'].append(loss)
            self.epoch_losses.append(loss)
            result['loss'] = loss

        # 处理 before_metric 激活
        if self.before_metric is None:
            y_hat_activated = y_hat
        elif isinstance(self.before_metric, dict):
            # 如果是字典类型，需要对不同的指标使用不同的激活函数
            y_hat_activated = {}
            for metric_name in self.metrics.keys():
                # 检查是否有针对特定指标的激活函数
                if metric_name in self.before_metric:
                    activation_fn = self.before_metric[metric_name]
                # 检查是否有通配符激活函数
                elif '*' in self.before_metric:
                    activation_fn = self.before_metric['*']
                else:
                    activation_fn = None

                if activation_fn is not None:
                    y_hat_activated[metric_name] = activation_fn(y_hat)
                else:
                    y_hat_activated[metric_name] = y_hat
        else:
            # 如果是单个激活函数
            if self.before_metric == torch.softmax:
                # softmax 需要指定 dim 参数
                y_hat_activated = self.before_metric(y_hat, dim=-1)
            else:
                y_hat_activated = self.before_metric(y_hat)

        # Compute and log metrics
        for metric_name, metric in self.metrics.items():
            # 如果 y_hat_activated 是字典，使用对应的激活结果
            if isinstance(y_hat_activated, dict):
                current_y_hat = y_hat_activated[metric_name]
            else:
                current_y_hat = y_hat_activated

            step_metric_value = metric(current_y_hat, y)
            result[metric_name] = step_metric_value.item()
            self.step_history[metric_name].append(step_metric_value.item())
            # Update epoch-level metrics
            self.epoch_metrics[metric_name].update(current_y_hat, y)

        self.current_step += 1
        return result

    def compute(self):
        """计算一个 epoch 的评估指标; 在每个 epoch 结束时调用"""
        epoch_metrics = {}

        # Compute epoch-level loss
        if self.epoch_losses:
            epoch_loss = sum(self.epoch_losses) / len(self.epoch_losses)
            epoch_metrics['loss'] = epoch_loss
            # self.fn_logging('epoch_loss', epoch_loss)

        # Compute and log epoch-level metrics
        for metric_name, metric in self.epoch_metrics.items():
            epoch_metric_value = metric.compute()
            epoch_metrics[metric_name] = epoch_metric_value.item()
            # self.fn_logging(f'epoch_{metric_name}', epoch_metric_value.item())
            metric.reset()

        self.epoch_history.append(epoch_metrics)
        self.current_epoch += 1
        self.current_step = 0
        self.epoch_losses = []
        return epoch_metrics

    def get_history(
        self, metric_name: str, level: Literal['step', 'epoch'] = 'step'
    ) -> List[float]:
        if level == 'step':
            return self.step_history[metric_name]
        elif level == 'epoch':
            return [
                epoch[metric_name] for epoch in self.epoch_history if metric_name in epoch
            ]
        else:
            raise ValueError("Level must be either 'step' or 'epoch'")



from zyplib.signal.psd import (
    bandpower,
    eeg_alpha_power,
    eeg_beta_power,
    eeg_delta_power,
    eeg_sigma_power,
    eeg_theta_power,
    welch,
)

__all__ = [
    'bandpower',
    'eeg_alpha_power',
    'eeg_beta_power',
    'eeg_delta_power',
    'eeg_sigma_power',
    'eeg_theta_power',
    'welch',
]



from functools import partial
from typing import Callable, Dict, Literal, Optional, Union

import lightning as L
import torch
from torch import nn
from torch.optim.lr_scheduler import _LRScheduler
from torchmetrics import Metric, MetricCollection

from zyplib.train.metric.torch_metric import TorchMetricRecorder
from zyplib.train.utils import prefixed_dict


class BasicLightningTrainer(L.LightningModule):
    def __init__(
        self,
        model: nn.Module,
        optimizer: torch.optim.Optimizer,
        fn_loss: nn.Module = None,
        metrics: Optional[Union[MetricCollection, Dict[str, Metric]]] = None,
        before_metric: Optional[
            Union[Literal['sigmoid', 'softmax'], Callable, nn.Module]
        ] = None,
        lr_scheduler: Optional[_LRScheduler] = None,
        lr_sched_freq: int = 1,
        lr_sched_interval: str = 'epoch',
    ):
        """一个基础常用的 LightningModule Trainer

        Parameters
        ----------
        - `model` : `nn.Module`
            - 模型
        - `optimizer` : `torch.optim.Optimizer`
            - 优化器
        - `fn_loss` : `nn.Module`, optional
            - 损失函数
        - `metrics` : `MetricCollection | Dict[str, Metric]`, optional
            - 评估指标
            - 使用 `torchmetrics.MetricCollection` 或 `dict` 类型
        - `before_metric` : `'sigmoid' | 'softmax' | Callable | nn.Module`, optional
            - 在计算评估指标之前的激活函数
            - 例如在二分类问题中，模型最后一层没有激活函数，而在计算评估指标 (即 `metrics`) 之前需要使用 `torch.sigmoid` 激活
        - `lr_scheduler` : `torch.optim.lr_scheduler._LRScheduler`, optional
            - 学习率调度器
        - `lr_sched_freq` : `int`, optional
            - 学习率调度器更新频率
        - `lr_sched_interval` : `str`, optional
            - 学习率调度器更新间隔
        """
        L.LightningModule.__init__(self)

        self.train_metric = TorchMetricRecorder(metrics, before_metric=before_metric)
        self.val_metric = TorchMetricRecorder(metrics, before_metric=before_metric)
        self.test_metric = TorchMetricRecorder(metrics, before_metric=before_metric)

        self.model = model
        self.optimizer = optimizer
        self.fn_loss = fn_loss or nn.BCEWithLogitsLoss()
        self.lr_scheduler = lr_scheduler
        self.lr_sched_freq = lr_sched_freq
        self.lr_sched_interval = lr_sched_interval

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self.forward(x)
        loss = self.fn_loss(y_hat, y)
        result = self.train_metric.update(y_hat, y, loss=loss.item())
        self.log_dict(prefixed_dict('train_', result), prog_bar=True, on_epoch=True, on_step=False)
        return loss

    def on_train_epoch_end(self) -> None:
        self.train_metric.compute()
        # print('train_result', result)

    def validation_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self.forward(x)
        loss = self.fn_loss(y_hat, y)
        result = self.val_metric.update(y_hat, y, loss=loss.item())
        self.log_dict(prefixed_dict('val_', result), prog_bar=True, on_epoch=True, on_step=False)
        return loss

    def on_validation_epoch_end(self) -> None:
        self.val_metric.compute()
        # print('val_result', result)

    def test_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self.forward(x)
        loss = self.fn_loss(y_hat, y)
        result = self.test_metric.update(y_hat, y, loss=loss.item())
        self.log_dict(prefixed_dict('test_', result), prog_bar=True, on_epoch=True, on_step=False)
        return loss

    def on_test_epoch_end(self) -> None:
        self.test_metric.compute()
        # print('test_result', result)

    def configure_optimizers(self):
        if self.lr_scheduler is None:
            return self.optimizer
        return {
            'optimizer': self.optimizer,
            'lr_scheduler': {
                'scheduler': self.lr_scheduler,
                'interval': self.lr_sched_interval,
                'frequency': self.lr_sched_freq,
                'monitor': 'val_loss',
            },
        }



from functools import singledispatch
from typing import Literal, Optional, Union

import numpy as np
import torch

__all__ = ['ensure_3dims', 'ensure_2dims', 'ensure_npy', 'ensure_tensor', 'ensure_scalar']


def numel(data: torch.Tensor | np.ndarray):
    if isinstance(data, torch.Tensor):
        return data.numel()
    elif isinstance(data, np.ndarray):
        return data.size
    else:
        raise TypeError(
            f'data should be torch.Tensor or np.ndarray, but got {type(data)}'
        )


def ensure_scalar(
    data: int | float | complex | torch.Tensor | np.ndarray | list,
) -> int | float | complex | None:
    """确保 data 是标量

    Parameters
    ----------
    - `data` : 输入的数据

    Returns
    ----------
    - `int | float | complex | None`
        - 将输入的数据转换为标量; 如果无法转换为标量，则返回 None
    """
    if isinstance(data, (int, float, complex)):
        return data
    elif isinstance(data, (torch.Tensor, np.ndarray)):
        data = data.squeeze()
        if numel(data) == 1:
            return data.item()
    elif isinstance(data, (list, tuple)):
        if len(data) == 1:
            return data[0]
    return None


def ensure_npy(
    data: Union[np.ndarray, torch.Tensor, list], dtype: Optional[np.dtype] = None
):
    """确保 data 是 np.ndarray 格式

    Parameters
    ----------
    - `data` : Union[np.ndarray, torch.Tensor, list]
        - 输入的数据

    Returns
    ----------
    - `np.ndarray`
        - 将输入的数据转换为 np.ndarray 格式
    """
    if isinstance(data, np.ndarray):
        ...
    elif isinstance(data, torch.Tensor):
        data = data.detach().cpu().numpy()
    elif isinstance(data, list):
        data = np.array(data)
    else:
        raise TypeError(
            f'signals should be np.ndarray or torch.Tensor, but got {type(data)}'
        )
    if dtype is not None:
        data = data.astype(dtype)
    return data


def ensure_tensor(
    data: Union[np.ndarray, torch.Tensor, list],
    dtype: Optional[torch.dtype] = None,
    detach: bool = False,
    device: Optional[Union[str, torch.device]] = None,
):
    """确保 data 是 torch.Tensor 格式

    Parameters
    ----------
    - `data` : Union[np.ndarray, torch.Tensor, list]
        - 输入的数据
    - `dtype` : Optional[torch.dtype]
        - 数据类型
    - `detach` : bool
        - 是否要 detach，默认 true
    - `device` : `torch.device | str`
        - 设备，默认 none，表示不转换 device

    Returns
    ----------
    - `torch.Tensor`
        - 将输入的数据转换为 torch.Tensor 格式
    """
    if isinstance(data, np.ndarray):
        data = torch.from_numpy(data)
    elif isinstance(data, torch.Tensor):
        ...
    elif isinstance(data, list):
        data = torch.tensor(data)
    else:
        raise TypeError(
            f'signals should be np.ndarray or torch.Tensor, but got {type(data)}'
        )
    if dtype is not None:
        data = data.type(dtype)
    if detach:
        data = data.detach()
    if device is not None:
        data = data.to(device)
    return data


@singledispatch
def ensure_3dims(
    signals: Union[np.ndarray, torch.Tensor],
    newaxis: Literal['batch', 'channel'] = 'channel',
):
    """把数据转换为标准的 [Batch, Channel, T] 格式

    ```py
    X = np.random.randn(2, 3)
    X1 = ensure_3dims(X, newaxis='channel')
    print(X1.shape)  # [2, 1, 3]
    X2 = ensure_3dims(X, newaxis='batch')
    print(X2.shape)  # [1, 2, 3]
    ```

    Parameters:
    ----------
    - `signals`: np.ndarray | torch.Tensor, 信号张量, shape 可以为 [T] 或者 [N, T]
    - `newaxis`: Literal['batch', 'channel'], 指定在哪个维度上添加新的维度
        - 'batch': 在 batch 维度上添加新的维度
        - 'channel': 在 channel 维度上添加新的维度
    """
    ...


@ensure_3dims.register(np.ndarray)
def _ensure_3dims_npy(
    signals: np.ndarray, newaxis: Literal['batch', 'channel'] = 'channel'
):
    if signals.ndim != 3:
        if signals.ndim == 1:
            signals = signals[np.newaxis, np.newaxis, :]
        elif signals.ndim == 2:
            if newaxis == 'channel':
                signals = signals[:, np.newaxis, :]
            else:
                signals = signals[np.newaxis, :, :]
        else:
            raise ValueError(f'signals.ndim should be 1 or 2, but got {signals.ndim}')
    return signals


@ensure_3dims.register(torch.Tensor)
def _ensure_3dims_torch(
    signals: torch.Tensor, newaxis: Literal['batch', 'channel'] = 'channel'
):
    if signals.ndim != 3:
        if signals.ndim == 1:
            signals = signals.unsqueeze(0).unsqueeze(0)
        elif signals.ndim == 2:
            if newaxis == 'channel':
                signals = signals.unsqueeze(1)
            else:
                signals = signals.unsqueeze(0)
        else:
            raise ValueError(f'signals.ndim should be 1 or 2, but got {signals.ndim}')
    return signals


@singledispatch
def ensure_2dims(signals: Union[np.ndarray, torch.Tensor]):
    """把数据转换为标准的 [Batch, T] 格式

    Parameters:
    ----------
    - `signals`: np.ndarray | torch.Tensor, 信号张量, shape 可以为 [T] 或者 [N, T]
    """
    ...


@ensure_2dims.register(np.ndarray)
def _ensure_2dims_npy(signals: np.ndarray):
    if signals.ndim != 2:
        if signals.ndim == 1:
            signals = signals[np.newaxis, :]
        else:
            signals = signals.squeeze()
            if signals.ndim == 2:
                ...
            elif signals.ndim == 1:
                signals = signals[np.newaxis, :]
            else:
                raise ValueError(f'signals.ndim should be 1 or 3, but got {signals.ndim}')
    return signals


@ensure_2dims.register(torch.Tensor)
def _ensure_2dims_torch(signals: torch.Tensor):
    if signals.ndim != 2:
        if signals.ndim == 1:
            signals = signals.unsqueeze(0)
        else:
            signals = signals.squeeze()
            if signals.ndim == 2:
                ...
            elif signals.ndim == 1:
                signals = signals.unsqueeze(0)
            else:
                raise ValueError(f'signals.ndim should be 1 or 3, but got {signals.ndim}')
    return signals



from typing import Callable, Literal

import matplotlib.pyplot as plt
import numpy as np

from zyplib.annotation import Event


def plot_signals(
    signals: np.ndarray,
    srate: int,
    labels: list[str] | None = None,
    new_fig: bool = True,
    show=True,
    save_path=None,
    interval: float | Literal['auto'] | None = 'auto',
    fn_value_fmt: Callable[[float], str] | None = lambda x: f'{x:.2f}',
    cmap: str | Callable[[int], str] | None = 'black',
    t_start: float = 0,
    events: list[Event] | None = None,
    xlabel: str = 'Time (s)',
    title: str = 'Multi-channel Signal Plot',
    show_ch_range=True
):
    """绘制多通道信号波形图

    Parameters
    ----------
    - `signals` : `np.ndarray`, [n_channels, n_samples]; 信号数据
    - `labels` : `list[str]`; 每个信号的标签
    - `srate` : `int`; 采样率
    - `new_fig` : `bool`, optional
        - 是否创建一个新的 figure。默认为 True
    - `show` : `bool`, optional
        - 是否显示绘图。默认为 True
    - `save_path` : `str`, optional
        - 保存绘图的路径。如果为 None，则不保存绘图。默认为 None
    - `interval` : `float | Literal['auto']`, optional
        - 不同通道在 Y 轴上的间隔, 默认 'auto' 表示自动计算
    - `fn_value_fmt` : `Callable[[float], str]`, optional
        - 用于格式化信号值的函数。默认为 `lambda x: f'{x:.2f}'`
    - `cmap` : `str | Callable[[int], str] | None`, optional
        - 用于绘制信号的 colormap。默认为 'black'
        - 也可以为一个 `int -> str` 的函数，int 为通道索引
    - `t_start` : `float`, optional
        - 信号的起始时间。默认为 0
    - `events` : `list[Event] | None`, optional
        - 标注事件。默认为 None
    """
    n_channels, n_samples = signals.shape
    t_start = max(t_start, 0)
    t_end = t_start + n_samples / srate
    time = np.linspace(t_start, t_end, n_samples)

    labels = labels or [f'Ch{i}' for i in range(n_channels)]

    # 计算间隔
    if interval is None or interval == 'auto':
        interval = 1.5 * np.max(np.abs(signals))

    # detrends
    signals = np.apply_along_axis(lambda x: x - np.mean(x), 1, signals)
    y_centers = np.arange(n_channels) * interval

    if new_fig:
        # fig = plt.figure(figsize=(15, n_channels * 2))
        fig = plt.figure()
    else:
        fig = plt.gcf()
    for i in range(n_channels):
        c = cmap(i) if callable(cmap) else cmap
        plt.plot(time, signals[i] + y_centers[i], label=labels[i], color=c)

    plt.xlim(t_start, t_end)
    if xlabel:
        plt.xlabel(xlabel)
    if title:
        plt.title(title)
    plt.grid(True)

    # 左侧标注通道名称
    plt.yticks(y_centers, labels)

    # Y 轴右侧标注 signal 的 min/max
    if show_ch_range:
        ax = plt.gca()
        ax2 = ax.twinx()
        ax2.set_ylim(ax.get_ylim())
        ax2.set_yticks(y_centers)
        min_max_labels = [
            f'{fn_value_fmt(signals[i].min())}~{fn_value_fmt(signals[i].max())}'
            for i in range(n_channels)
        ]
        ax2.set_yticklabels(min_max_labels)
        # ax2.set_ylabel('Min / Max')

    # 绘制标注
    if events:
        y_lim = ax.get_ylim()
        y_mid = (y_lim[0] + y_lim[1]) / 2
        for event in events:
            onset = event.onset
            duration = event.duration
            label = event.name
            if duration is None:
                plt.axvline(onset, color='red', linestyle='--', label=label)
            else:
                plt.axvspan(onset, onset + duration, color='red', alpha=0.3, label=label)
            if label:
                plt.text(
                    onset,
                    y_mid,  # 显示在 y 轴的中间位置
                    label,
                    color='red',
                    fontsize=10,
                    verticalalignment='center',
                    horizontalalignment='right',
                )

    plt.tight_layout()

    if save_path:
        plt.savefig(save_path)

    if show:
        plt.show()

    return fig



import numpy as np
from scipy import signal

from zyplib.utils.ensure import ensure_npy

__all__ = [
    'welch',
    'bandpower',
    'eeg_delta_power',
    'eeg_theta_power',
    'eeg_alpha_power',
    'eeg_sigma_power',
    'eeg_beta_power',
]


def welch(x, fs, nperseg=None, max_freq=None, axis=-1):
    """Welch PSD
    计算 welch 功率谱, 最后截断结果, 返回低于 lowpass Hz 的部分

    Parameters
    ----------
    - x : array_like
    - fs : int, 采样率
    - nperseg : int, 每段的长度, 默认 2 * fs, 也就意味着频率分辨率 0.5 Hz
    - max_freq : int, 截断频率
        - 当非 None 的时候, 截断超过 max_freq 的部分
    - axis : int, 默认 -1,即最后一个维度

    Returns
    ----------
    - f : ndarray, 频率
        - 分辨率为 fs / nperseg
        - 长度为 nperseg / 2
    - Pxx : ndarray, 功率谱
    """
    x = ensure_npy(x)
    if not nperseg:
        nperseg = fs * 2
    nfft = np.power(2, np.ceil(np.log2(nperseg)))
    f, Pxx = signal.welch(x, fs, nperseg=nperseg, nfft=nfft, axis=axis)
    # 当需要截断且截断频率小于 fs / 2 时, 进行截断
    if max_freq is not None and max_freq < fs / 2:
        ind = np.argmax(f > max_freq)
        if ind == 0:
            raise ValueError('lowpass is too low')
        f = f[:ind]
        Pxx = Pxx[..., :ind]
    return f, Pxx


def bandpower(x, fs, band: tuple, axis=-1, nperseg=None):
    x = ensure_npy(x)
    f, Pxx = welch(x, fs, axis=axis, nperseg=nperseg)
    ind_min = np.argmax(f > band[0])
    ind_max = np.argmax(f > band[1])
    if ind_min == 0 or ind_max == 0:
        raise ValueError('band values are too low')
    return np.trapz(Pxx[..., ind_min:ind_max], f[ind_min:ind_max], axis=axis)


def eeg_delta_power(x, fs, axis=-1, nperseg=None):
    """Delte 0.5 and 4 Hz.

    Parameters
    ----------
    - `x` : array_like, 1-D or 2-D
        - 待处理的原始 EEG 信号, 可以是一维或二维数组, 类型为 `array_like`。
    - `fs` : float
        - 输入信号的采样频率, 单位为 Hz, 类型为 `float`。
    - `axis` : int, optional
        - 可选参数, 计算取样信号频谱的轴。默认值为 -1, 即最后一个轴, 类型为 `int`。

    Returns
    ----------
    - `band_power` : float
    """

    return bandpower(x, fs, (0.5, 4), axis=axis, nperseg=nperseg)


def eeg_theta_power(x, fs, axis=-1, nperseg=None):
    """Theta 4 and 8 Hz.

    Parameters
    ----------
    - `x` : array_like, 1-D or 2-D
        - 待处理的原始 EEG 信号, 可以是一维或二维数组, 类型为 `array_like`。
    - `fs` : float
        - 输入信号的采样频率, 单位为 Hz, 类型为 `float`。
    - `axis` : int, optional
        - 可选参数, 计算取样信号频谱的轴。默认值为 -1, 即最后一个轴, 类型为 `int`。

    Returns
    ----------
    - `band_power` : float
    """
    return bandpower(x, fs, (4, 8), axis=axis, nperseg=nperseg)


def eeg_alpha_power(x, fs, axis=-1, nperseg=None):
    """Alpha 8 and 12 Hz.

    Parameters
    ----------
    - `x` : array_like, 1-D or 2-D
        - 待处理的原始 EEG 信号, 可以是一维或二维数组, 类型为 `array_like`。
    - `fs` : float
        - 输入信号的采样频率, 单位为 Hz, 类型为 `float`。
    - `axis` : int, optional
        - 可选参数, 计算取样信号频谱的轴。默认值为 -1, 即最后一个轴, 类型为 `int`。

    Returns
    ----------
    - `band_power` : float
    """
    return bandpower(x, fs, (8, 12), axis=axis, nperseg=nperseg)


def eeg_sigma_power(x, fs, axis=-1, nperseg=None):
    """Sigma 12 and 15 Hz.

    Parameters
    ----------
    - `x` : array_like, 1-D or 2-D
        - 待处理的原始 EEG 信号, 可以是一维或二维数组, 类型为 `array_like`。
    - `fs` : float
        - 输入信号的采样频率, 单位为 Hz, 类型为 `float`。
    - `axis` : int, optional
        - 可选参数, 计算取样信号频谱的轴。默认值为 -1, 即最后一个轴, 类型为 `int`。

    Returns
    ----------
    - `band_power` : float
    """
    return bandpower(x, fs, (12, 15), axis=axis, nperseg=nperseg)


def eeg_beta_power(x, fs, axis=-1, nperseg=None):
    """Beta 12 and 30 Hz.

    Parameters
    ----------
    - `x` : array_like, 1-D or 2-D
        - 待处理的原始 EEG 信号, 可以是一维或二维数组, 类型为 `array_like`。
    - `fs` : float
        - 输入信号的采样频率, 单位为 Hz, 类型为 `float`。
    - `axis` : int, optional
        - 可选参数, 计算取样信号频谱的轴。默认值为 -1, 即最后一个轴, 类型为 `int`。

    Returns
    ----------
    - `band_power` : float
    """
    return bandpower(x, fs, (12, 30), axis=axis, nperseg=nperseg)



from typing import Literal, Union

import numpy as np
from scipy import fft, signal

from zyplib.utils.ensure import ensure_npy

# FilterType = Literal['butter', 'cheby1', 'cheby2', 'ellip', 'bessel']
__all__ = ['notch', 'bandpass_butter']


def notch(X, fs, f0=50, Q=30, axis=-1):
    r"""Notch filter.

    Parameters
    ----------
    - signals : array_like
        - Input signal.
    - fs : float
        - Sampling frequency.
    - f0 : float
        - Notch frequency.
    - Q : float, optional
        - Q factor. Default is 30.
        - https://en.wikipedia.org/wiki/Q_factor
        - $Q = \frac{f_0}{\Delta f}$
    - axis : int, optional
        - Axis along which to apply the filter. Default is -1.

    Returns
    ----------
    - y : ndarray
        - Filtered signal.
    """
    X = ensure_npy(X)
    b, a = signal.iirnotch(f0, Q, fs)
    return signal.lfilter(b, a, X, axis=axis)


def bandpass_butter(
    X, fs: float, lfreq: float, rfreq: float, order=3, prevent_nan=True, axis=-1
):
    r"""Bandpass filter.

    Parameters
    ----------
    - signals : array_like
        - Input signal.
    - fs : float
        - Sampling frequency.
    - lfreq : float
        - Left frequency.
    - rfreq : float
        - Right frequency.
    - order : int, optional
        - Filter order. Default is 3.
        - 注意, 太高的阶数（例如 5）可能会导致滤波器不稳定，使得结果出现 NaN 值
    - prevent_nan : bool, optional
        - 如果为 True，则尝试降低滤波器阶数以防止 NaN 值。
    - axis : int, optional

    Returns
    ----------
    - y : ndarray
        - Filtered signal.
    """
    X = ensure_npy(X)
    nyq = 0.5 * fs
    low = lfreq / nyq
    high = rfreq / nyq

    # 检查频率范围是否有效
    if low >= high or high >= 1:
        raise ValueError('Invalid frequency range. Ensure lfreq < rfreq < fs/2')

    b, a = signal.butter(order, [low, high], btype='bandpass')
    y = signal.lfilter(b, a, X, axis=axis)

    # 检查输出是否包含 NaN 值
    if np.isnan(y).any():
        print('Warning: NaN values detected in the output. Trying with a lower order.')

        if prevent_nan:
            return bandpass_butter(X, fs, lfreq, rfreq, order=max(1, order - 1), axis=axis)
        else:
            raise ValueError('NaN values detected in the output.')

    return y



"""
同构建模型训练数据集相关的模块
"""

from .labels import from_onehot, to_onehot
from .make_xy_pipeline import basic_pipeline, kfold_pipeline, train_test_pipeline
from .split import random_split
from .torch import BasicXyDataset as TorchBasicXyDataset
from .torch import count_dataloader_labels, make_basic_dataloader, make_dataloader_from_xy

__all__ = [
    'basic_pipeline',
    'kfold_pipeline',
    'train_test_pipeline',
    'random_split',
    'TorchBasicXyDataset',
    'make_basic_dataloader',
    'count_dataloader_labels',
    'from_onehot',
    'to_onehot',
]



from torch import nn
from torch.nn.utils import weight_norm

from zyplib.nn import BaseModule


class CausalConv1d(nn.Conv1d):
    """1D DILATED CAUSAL CONVOLUTION."""

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        dilation=1,
        bias=True,
        padding_mode='zeros',
    ):
        super(CausalConv1d, self).__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=(kernel_size - 1) * dilation,
            dilation=dilation,
            bias=bias,
            padding_mode=padding_mode,
        )
        self.pad_size = (kernel_size - 1) * dilation

    def forward(self, x):
        y = super(CausalConv1d, self).forward(x)
        if self.pad_size != 0:
            y = y[:, :, : -self.pad_size].contiguous()
        return y


class TemporalBlock(nn.Module):
    def __init__(
        self,
        in_cn,
        out_cn,
        kernel_size,
        stride,
        dilation,
        dropout=0.2,
        padding_mode='zeros',
        activation=nn.ReLU,
    ):
        super(TemporalBlock, self).__init__()
        self.conv1 = weight_norm(
            CausalConv1d(
                in_cn,
                out_cn,
                kernel_size,
                stride=stride,
                dilation=dilation,
                padding_mode=padding_mode,
            )
        )
        self.relu1 = activation()
        self.dropout1 = nn.Dropout(dropout)

        self.conv2 = weight_norm(
            CausalConv1d(
                in_cn, out_cn, kernel_size, dilation=dilation, padding_mode=padding_mode
            )
        )
        self.relu2 = activation()
        self.dropout2 = nn.Dropout(dropout)

        self.net = nn.Sequential(
            self.conv1, self.relu1, self.dropout1, self.conv2, self.relu2, self.dropout2
        )
        self.downsample = nn.Conv1d(in_cn, out_cn, 1) if in_cn != out_cn else None
        self.relu = activation()

    def forward(self, x):
        out = self.net(x)
        res = x if self.downsample is None else self.downsample(x)
        return self.relu(out + res)


class TemporalConvNet(BaseModule):
    def __init__(
        self,
        in_ch: int,
        level_ch: list,
        num_stack=1,
        kernel_size=2,
        dropout=0.2,
        padding_mode='zeros',
    ):
        super(TemporalConvNet, self).__init__()
        layers = []
        num_levels = len(level_ch)
        dilations = [2**i for i in range(num_levels)] * num_stack
        level_ch *= num_stack
        self.reciptive_field_size = sum(dilations) * (kernel_size - 1) + 1

        for dilation_size, out_ch in zip(dilations, level_ch):
            layers += [
                TemporalBlock(
                    in_ch,
                    out_ch,
                    kernel_size,
                    stride=1,
                    dilation=dilation_size,
                    dropout=dropout,
                    padding_mode=padding_mode,
                )
            ]
            in_ch = out_ch

        self.network = nn.Sequential(*layers)

    def forward(self, x):
        return self.network(x)



from typing import Any, Dict, Union

import torch
import torch.optim
from torch import nn

from zyplib.utils.time import now

from .shapes import output_shape

__all__ = [
    'output_shape',
    'weight_initializer',
    'save_checkpoint',
    'load_checkpoint',
    'freeze',
    'unfreeze',
]


def weight_initializer(m: nn.Module):
    if isinstance(m, (nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.Linear)):
        nn.init.kaiming_normal_(m.weight.data, mode='fan_out', nonlinearity='relu')
        if m.bias is not None:
            nn.init.constant_(m.bias.data, 0)
    elif isinstance(m, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d)):
        nn.init.constant_(m.weight.data, 1)
        nn.init.constant_(m.bias.data, 0)
    elif isinstance(m, nn.LSTM):
        for name, param in m.named_parameters():
            if 'weight' in name:
                nn.init.orthogonal_(param.data)
            elif 'bias' in name:
                nn.init.constant_(param.data, 0)
    elif isinstance(m, nn.GRU):
        for name, param in m.named_parameters():
            if 'weight' in name:
                nn.init.xavier_uniform_(param.data)
            elif 'bias' in name:
                nn.init.constant_(param.data, 0)
    elif isinstance(m, nn.MultiheadAttention):
        for param in m.parameters():
            if param.dim() > 1:
                nn.init.xavier_uniform_(param)


def save_checkpoint(
    models: Union[Dict[str, nn.Module], nn.Module],
    path: str,
    optimizers: Union[Dict[str, Any], Any] = None,
    epoch=None,
    loss=None,
    **kwargs,
):
    if isinstance(loss, torch.Tensor):
        loss = loss.item()

    if not isinstance(models, dict):
        models = {'model': models}
    if optimizers is not None and not isinstance(optimizers, dict):
        optimizers = {'optimizer': optimizers}

    checkpoint = {
        'models_state_dict': {name: model.state_dict() for name, model in models.items()},
        'optimizers_state_dict': {
            name: opt.state_dict() for name, opt in optimizers.items()
        }
        if optimizers
        else None,
        'epoch': epoch,
        'loss': loss,
        'save_time': now(),
        **kwargs,
    }
    checkpoint = {k: v for k, v in checkpoint.items() if v is not None}
    torch.save(checkpoint, path)


def load_checkpoint(
    models: Union[Dict[str, nn.Module], nn.Module],
    path: str,
    optimizers: Union[Dict[str, Any], Any] = None,
    device: Union[str, torch.device] = None,
) -> Dict[str, Any]:
    checkpoint = torch.load(path, map_location=device)

    if not isinstance(models, dict):
        models = {'model': models}
    if optimizers is not None and not isinstance(optimizers, dict):
        optimizers = {'optimizer': optimizers}

    for name, model in models.items():
        model.load_state_dict(checkpoint['models_state_dict'][name])

    if optimizers and 'optimizers_state_dict' in checkpoint:
        for name, optimizer in optimizers.items():
            if name in checkpoint['optimizers_state_dict']:
                optimizer.load_state_dict(checkpoint['optimizers_state_dict'][name])

    return checkpoint


def freeze(model: nn.Module):
    for param in model.parameters():
        param.requires_grad = False


def unfreeze(model: nn.Module):
    for param in model.parameters():
        param.requires_grad = True



import argparse
import sys

from zyplib.utils.cli import parse_cli_args
from zyplib.utils.fs import write_json
from zyplib.utils.print import print_error, print_info, print_warning

commands = {}
kv_args = {}


def add_command(name):
    def decorator(func):
        commands[name] = func
        return func

    return decorator


@add_command('new-config-file')
def create_config_file():
    """创建一个 zyplib.config.json 文件"""
    from . import _config

    fname = 'zyplib.config.json'
    write_json(fname, _config.config.asdict())


@add_command('show-config')
def show_config():
    """显示当前配置"""
    from . import _config

    print_info(_config.config.asdict())


@add_command('reset-global-config')
def reset_global_config():
    """重置全局配置"""
    from . import _config

    config = _config.Config()
    print_info(f'重置全局配置为: {config}')
    _config._write_config(config)


@add_command('show-caches')
def show_caches():
    """显示当前缓存, 可选参数 table={true, false}"""
    from zyplib.utils.cache import display_caches

    show = kv_args.get('table', True)
    display_caches(table=show)


@add_command('viz-eeg')
def visualize_and_annotate():
    """可视化脑电信号并标注, 可选参数 path=str, srate=int"""
    from zyplib.apps.annotators import launch_app

    path = kv_args.get('path')
    srate = kv_args.get('srate')
    launch_app(fpath=path, srate=srate)


@add_command('inspect-keyval')
def inspect_key_value():
    """在目录中递归查看所有 YAML/JSON 等 key-value 文件中指定 key 名称的取值，以表格的形式展示
    参数: dir=str, key=str
    用例: python -m zyplib.cli inspect-keyval dir="./results" key="score"
    """

    from prettytable import PrettyTable

    from zyplib.utils.kv import collect_key_path

    dir: str = kv_args.get('dir')
    key: str = kv_args.get('key')
    # 搜索文件并获取结果
    results = collect_key_path(dir, key)

    # 展示结果
    for filename, key_paths in results.items():
        table = PrettyTable(['Path', '取值'])
        for k, v in key_paths.items():
            table.add_row([k, v])

        if len(table.rows) > 0:
            print_info(f'{filename}'.center(100, '='))
            print(table)


def help_doc():
    docs = []
    for command, func in commands.items():
        docs.append(f'* {command}: {func.__doc__}')
    return '\n'.join(docs)


def main():
    parser = argparse.ArgumentParser(
        description='ZypLib CLI tool',
        epilog=help_doc(),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('command', type=str, help='Command to run', nargs='?')
    parser.add_argument('args', nargs=argparse.REMAINDER, help='Additional arguments')

    args = parser.parse_args()

    global kv_args
    kv_args = parse_cli_args()

    command = getattr(args, 'command', None)
    if command and command in commands:
        commands[command]()
    else:
        print_warning(f'未知的命令: {command}')
        print(help_doc())

    sys.exit(0)


if __name__ == '__main__':
    main()



import os
from typing import Callable, Optional

from lightning.fabric import Fabric, loggers

from zyplib.utils.time import TIME_FMT, now

from . import trainer
from .callbacks import (
    FabricCallbackProtocol,
    ParseCSVLogfile,
    PrintTrainInfo,
    SaveCheckpoints,
    SaveHparams,
    find_callback,
)
from .procedure import (
    FabricTrainingContext,
    train_epoch,
    train_loop,
    training_step,
    val_epoch,
    validation_step,
)
from .trainer import BasicFabricTrainer

basic_trainer = trainer  # 兼容旧的模块命名

__all__ = [
    'BasicFabricTrainer',
    'FabricTrainingContext',
    'default_fabric',
    'train_epoch',
    'train_loop',
    'training_step',
    'val_epoch',
    'validation_step',
    'find_callback',
]


def default_fabric(
    root_dir: str = './',
    dir_name: str = 'lightning_logs',
    version: str | None = None,
    save_every_k: Optional[int] = None,
    save_best: bool = True,
    keep_best_k: Optional[int] = 1,
    monitor: str = 'loss',
    less_is_better: bool = True,
    no_csv_logger: bool = False,
    no_tensorboard_logger: bool = False,
    no_print_info_cb: bool = False,
    no_save_checkpoints_cb: bool = False,
    no_save_hparams_cb: bool = False,
    fn_custom_logger: Optional[Callable[[str], list[loggers.Logger]]] = None,
    fn_custom_callback: Optional[Callable[[str], list[FabricCallbackProtocol]]] = None,
    **fabric_kwargs,
):
    """构建默认的fabric对象

    将 logger 和 checkpoint 保存在 `root_dir/dir_name/version` 目录下
    默认配置为`./lightning_logs/v2024_05_01-12_00_00`

    默认配置如下:
    - Loggers:
        - CSVLogger
        - TensorBoardLogger
    - Callbacks:
        - PrintTrainInfo
        - SaveHparams
        - SaveCheckpoints

    可以通过 `fn_custom_xxx` 参数来添加自定义的 logger 和 callback; 该回调函数传入 now 的时间字符串，需要返回一个 logger 或者 callback 的列表

    Parameters
    ----------
    - `root_dir` : `str`, optional
        - 日志保存的根目录, by default './'
    - `dir_name` : `str`, optional
        - 日志保存的目录名称, by default 'lightning_logs'
    - `version` : `str | None`, optional
        - 日志保存的版本, by default None
        - 默认为 `v{time_str}`
    - `save_every_k` : `Optional[int]`, SaveCheckpoints 的参数
        - 每多少个epoch保存一次模型, by default None
    - `save_best` : `bool`, SaveCheckpoints 的参数
        - _description_, by default True
    - `keep_best_k` : `Optional[int]`, SaveCheckpoints 的参数
        - 保存最好的k个模型, by default 1
    - `monitor` : `str`, SaveCheckpoints 的参数
        - 监控的指标, by default 'loss'
    - `less_is_better` : `bool`, SaveCheckpoints 的参数
        - 是否监控指标越小越好, by default True
    - `no_csv_logger` : `bool`, optional
        - 是否禁用CSVLogger, by default False
    - `no_tensorboard_logger` : `bool`, optional
        - 是否禁用TensorBoardLogger, by default False
    - `no_print_info_cb` : `bool`, optional
        - 是否禁用打印训练信息回调, by default False
    - `no_save_checkpoints_cb` : `bool`, optional
        - 是否禁用保存检查点回调, by default False
    - `no_save_hparams_cb` : `bool`, optional
        - 是否禁用保存超参数回调, by default False
    - `fn_custom_logger` : `(version: string) => Logger[]`, optional
        - 自定义 logger 的函数, by default None
    - `fn_custom_callback`: `(version: string) => FabricCallbackProtocol[]`
        - 自定义 callback 的函数, by default None

    Returns
    ----------
    - `fabric` : `Fabric`
        - Fabric 对象; 注意只返回 fabric，不自动调用 fabric.launch()
    """
    time_str = now(TIME_FMT.DATE_TIME_UNDERSCORE)
    version = version or f'v{time_str}'
    _loggers = []
    _callbacks = []

    # Add loggers if not disabled
    if not no_csv_logger:
        csv = loggers.CSVLogger(root_dir=root_dir, name=dir_name, version=version)
        _loggers.append(csv)
        parser = ParseCSVLogfile(
            os.path.join(csv.log_dir, 'metrics.csv'), to_excel=True, to_yaml=False
        )
        _callbacks.append(parser)
    if not no_tensorboard_logger:
        _loggers.append(
            loggers.TensorBoardLogger(root_dir=root_dir, name=dir_name, version=version)
        )

    if fn_custom_logger:
        custom_loggers = fn_custom_logger(version)
        if isinstance(custom_loggers, list):
            _loggers.extend(custom_loggers)
        else:
            raise ValueError(
                f'fn_custom_logger must return a list of loggers, but got {type(custom_loggers)}'
            )

    # Add callbacks if not disabled
    if not no_print_info_cb:
        _callbacks.append(PrintTrainInfo())
    if not no_save_hparams_cb:
        _callbacks.append(SaveHparams(dir=f'{root_dir}/{dir_name}/{version}'))
    if not no_save_checkpoints_cb:
        _callbacks.append(
            SaveCheckpoints(
                dir=f'{root_dir}/{dir_name}/{version}/checkpoints',
                save_every_k=save_every_k,
                save_best=save_best,
                keep_best_k=keep_best_k,
                monitor=monitor,
                less_is_better=less_is_better,
            )
        )
    if fn_custom_callback:
        custom_callbacks = fn_custom_callback(version)
        if isinstance(custom_callbacks, list):
            _callbacks.extend(custom_callbacks)
        else:
            raise ValueError(
                f'fn_custom_callback must return a list of callbacks, but got {type(custom_callbacks)}'
            )

    fabric = Fabric(loggers=_loggers, callbacks=_callbacks, **fabric_kwargs)
    return fabric



import re
from typing import List, Union


def match_channels_regex(
    src_channels: List[str], required_channels: List[Union[str, re.Pattern]]
):
    """基于正则表达式进行通道匹配通道

    从 src_channels 中按 required_channels 的顺序匹配出复合条件的通道名称

    ```
    results = []
    for ch_pattern in required_channels:
        从 src_channels 中匹配出符合 ch_pattern 的通道名称
        将匹配结果添加到 results 中
    return results
    ```

    Examples:
    ----------
    >>> data = StandardData.load_h5('./cache/UF4pbGI7/165974.edf.std-h5', meta_only=True)
    >>> required_channels = [r'F3-', r'F4-', r'C3-', r'C4-', r'O1-', r'O2-', r'(E1)|(EOG Left)', r'(E2)|(EOG Right)']
    >>> match_channels_regex(data.Channels, required_channels)

    Parameters
    ----------
        - `src_channels` : List[str]
        - `target_channels` : List[Regex Pattern | str]

    Returns
    ----------
        - `List[str]` 通道名称列表
        - `List[int]` 通道在 src_channels 中的索引
    """
    required_channels = [
        re.compile(ch) if isinstance(ch, str) else ch for ch in required_channels
    ]

    def match_channel(ch_pat: re.Pattern):
        for i, ch in enumerate(src_channels):
            if ch_pat.match(ch):
                return i
        return None

    matched_ch_names, matched_ch_index = [], []
    for ch_pat in required_channels:
        idx = match_channel(ch_pat)
        if idx is not None:
            matched_ch_names.append(src_channels[idx])
            matched_ch_index.append(idx)
    return matched_ch_names, matched_ch_index



import os
import re
from pathlib import Path

import numpy as np
import scipy.io as sio

from zyplib.annotation.event import Event

from ._base import EventMixin


class MatFile(EventMixin):
    DATA_PAT = r'^data(.*)'
    LABEL_PAT = r'^(channel|label)s?'
    SRATE_PAT = r'^(srate|sfreq|freq|sample_rate)s?'

    def __init__(
        self,
        filepath: str,
        srate: float | None = None,
        channels: list[str] | None = None,
        events: list[Event] | None = None,
        key_data: str | None = None,
        key_channels: str | None = None,
        key_srate: str | None = None,
        **kwargs,
    ):
        EventMixin.__init__(self, events)
        self.filepath = Path(filepath)

        self.data = None
        self.srate = srate
        self.channels = channels
        self.n_channels = 0
        self.n_samples = 0
        self.t_duration = 0

        self.name = os.path.basename(filepath)
        self.meta = {}

        try:
            mat = sio.loadmat(filepath)
            mat = {k: v for k, v in mat.items() if not k.startswith('__')}
            self._resolve_mat(mat, key_data, key_channels, key_srate)
        except Exception as e:
            raise RuntimeError(f'文件读取错误: {e}')

        self.n_samples = self.data.shape[1]
        self.t_duration = self.n_samples / self.srate

    def _resolve_mat(
        self,
        mat: dict,
        key_data: str | None,
        key_channels: str | None,
        key_srate: str | None,
    ):
        keys = list(mat.keys())

        # Use provided keys if available
        if key_data and key_data in mat:
            data = mat[key_data]
        else:
            for key in keys:
                if re.match(self.DATA_PAT, key):
                    data = mat[key]
                    break

        if key_channels and key_channels in mat:
            self.channels = mat[key_channels]
        else:
            for key in keys:
                if re.match(self.LABEL_PAT, key):
                    self.channels = mat[key]
                    break

        if key_srate and key_srate in mat:
            self.srate = mat[key_srate]
        else:
            for key in keys:
                if re.match(self.SRATE_PAT, key):
                    self.srate = mat[key]
                    break

        # =============== Section Beg: Data =============== #
        if data.ndim == 1:
            data = data.reshape(1, -1)
        # 数据默认应该是的第一维是通道数，第二维是时间
        CH, T = data.shape
        # 如果通道数大于时间，那么说明数据的第一维是时间，第二维是通道数
        if CH > T:
            data = data.T
            CH, T = data.shape
        self.data = data
        # --------------- Section End: Data --------------- #

        if self.channels is None:
            self.channels = [f'ch{i}' for i in range(CH)]

        if self.srate is None:
            raise RuntimeError('未找到采样率信息')

        self.n_channels = CH
        self.n_samples = T
        self.t_duration = T / self.srate

    def get_epoch(self, t_start: float, t_duration: float) -> np.ndarray:
        idx_start = int(t_start * self.srate)
        idx_end = idx_start + int(t_duration * self.srate)
        return self.data[:, idx_start:idx_end]

    @property
    def units(self):
        return { c: '?' for c in self.channels }



"""
- Background: 进行脑电等时序信号分析, 需要可视化信号、编辑标注信息等
- App: 基于 Tkinter 实现的信号可视化应用界面
- 功能:
    - 可视化信号波形图
    - 可视化标注信息
    - 新建标注
    - 删除标注
"""

import tkinter as tk
from dataclasses import dataclass
from pathlib import Path
from tkinter import messagebox, ttk
from typing import Callable

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

from zyplib.annotation import Event
from zyplib.io.interface import EEGFileInterface
from zyplib.utils.fs import write_json
from zyplib.utils.print import print_debug
from zyplib.vis import plot_signals


@dataclass
class AppState:
    t_start: float  # 开始时间, 单位: 秒
    t_page: float  # 页面长度, 单位: 秒
    interval: float | None
    active_event: Event | None = None


class TemporaryEventMixin:
    def __init__(self, *args, **kwargs):
        self.temporary_event = None
        self.line_2d = None
        self.span_2d = None

    def plot_temp_event(self, event: Event | None):
        self.temporary_event = event
        self.clear_temp_event()
        if event is None:
            return
        onset = event.onset
        duration = event.duration
        label = event.name
        if duration is None:
            self.line_2d = plt.axvline(onset, color='black', linestyle='--', label=label)
        else:
            self.span_2d = plt.axvspan(
                onset, onset + duration, color='black', alpha=0.2, label=label
            )
        plt.draw()  # 刷新绘图

    def clear_temp_event(self):
        self.temporary_event = None
        if self.line_2d:
            self.line_2d.remove()
            self.line_2d = None
        if self.span_2d:
            self.span_2d.remove()
            self.span_2d = None
        plt.draw()  # Redraw the plot to reflect the changes


class SignalAnnotatorApp(TemporaryEventMixin):
    def __init__(self, signals: EEGFileInterface, fn_on_exit: Callable[[], None]):
        super().__init__()

        self.signals = signals
        self.fn_on_exit = fn_on_exit
        self.state = AppState(t_start=0, t_page=4, interval=None, active_event=None)
        self.root = None
        self.canvas = None
        self.onset_entry = None
        self.duration_entry = None
        self.interval_entry = None

        self.annt_onset_entry = None
        self.annt_duration_entry = None
        self.annt_label_entry = None

    def on_click_figure(self, event):
        """鼠标点击波形图时的回调函数
        行为
        ----------
        - 鼠标左键:
            - 使用 `get_annotation` 方法, 获取在 `at.x` 左右 0.1 秒范围内的标注
            - 如果存在这样的标注, 记为 `state.active_event`
        - 鼠标右键: 在界面中添加新的标注
            - if active_event is None:
                - 在界面中添加新的标注, 更新 onset
                - 此时 active_event 的 onset 已知, duration 未知, 被视为一个新的单点标注
            - else if active_event is not None and active_event.duration is None
                - 更新 active_event.duration
                - 此时 active_event 的 onset 和 duration 都已知, 视为一个区域标注
            - else if active_event 是一个 区域标注
                - 清空 active_event
            - 做完之后, 更新 Figure
        """
        if not event.xdata:
            return
        if event.button == 1:  # 左键点击
            at_x = event.xdata
            annotations = self.signals.pick_event(at_x)
            if annotations:
                self.state.active_event = annotations[0]  # 选中第一个标注
                print_debug(f'选中标注: {self.state.active_event}')
            else:
                self.state.active_event = None
                print_debug('无标注')
            self.clear_temp_event()
            self.update_annotation_fields()
        elif event.button == 3:  # 右键点击
            at_x = event.xdata
            if self.state.active_event is None:
                # 创建新标注
                label_name = self.annt_label_entry.get() or ''
                self.state.active_event = Event(
                    name=label_name, onset=at_x, duration=None
                )
                self.update_annotation_fields()
                print_debug(f'创建新标注: {self.state.active_event}')
            else:
                if self.state.active_event.duration is None:
                    # 更新持续时间
                    self.state.active_event.duration = (
                        at_x - self.state.active_event.onset
                    )
                    print_debug(f'更新标注持续时间: {self.state.active_event}')
                else:
                    self.state.active_event = None  # 清除选中标注
                    print_debug('清除标注')
            self.update_annotation_fields()
            self.plot_temp_event(self.state.active_event)

    def update_plot(self):
        self.clear_temp_event()
        signals_data = self.signals.get_epoch(self.state.t_start, self.state.t_page)
        if self.state.interval is None or self.state.interval == 'auto':
            self.state.interval = 1.5 * np.max(np.abs(signals_data))
            self.update_entry_fields()

        fig = plt.gcf()
        # 清空 fig
        fig.clear()
        plot_signals(
            signals=signals_data,
            srate=int(self.signals.srate),
            labels=self.signals.channels,
            interval=self.state.interval,
            events=self.signals.get_events(self.state.t_start, self.state.t_page),
            t_start=self.state.t_start,
            new_fig=False,
            show=False,
            save_path=None,
        )
        self.canvas.figure = fig
        self.canvas.draw()

        self.state.active_event = None

    def update_entry_fields(self):
        self.onset_entry.delete(0, tk.END)
        self.onset_entry.insert(0, str(self.state.t_start))
        self.duration_entry.delete(0, tk.END)
        self.duration_entry.insert(0, str(self.state.t_page))

        if self.interval_entry:
            self.interval_entry.delete(0, tk.END)
            self.interval_entry.insert(0, str(self.state.interval))

    def update_annotation_fields(self):
        event = self.state.active_event
        if event is None:
            self.annt_onset_entry.delete(0, tk.END)
            self.annt_duration_entry.delete(0, tk.END)
            # self.annt_label_entry.delete(0, tk.END)
            return

        self.annt_onset_entry.delete(0, tk.END)
        self.annt_onset_entry.insert(0, str(event.onset))

        if event.duration is not None:
            self.annt_duration_entry.delete(0, tk.END)
            self.annt_duration_entry.insert(0, str(event.duration))

        if event.name:
            self.annt_label_entry.delete(0, tk.END)
            self.annt_label_entry.insert(0, event.name)

    def load_left(self):
        if self.state.t_start == 0:
            messagebox.showinfo('信息', '已到达左边界')
            return
        new_t_start = self.state.t_start - self.state.t_page
        if new_t_start < 0:
            new_t_start = 0
        self.state.t_start = new_t_start
        self.state.active_event = None
        self.update_entry_fields()
        self.update_annotation_fields()
        self.update_plot()

    def load_right(self):
        if self.state.t_start + self.state.t_page >= self.signals.t_duration:
            messagebox.showinfo('信息', '已到达右边界')
            return
        self.state.t_start += self.state.t_page
        if self.state.t_start + self.state.t_page > self.signals.t_duration:
            self.state.t_start = self.signals.t_duration - self.state.t_page
            messagebox.showinfo('信息', '已到达右边界')

        self.state.active_event = None
        self.update_entry_fields()
        self.update_annotation_fields()
        self.update_plot()

    def add_annotation(self):
        if self.state.active_event is None:
            return
        self.signals.add_event(self.state.active_event)
        self.state.active_event = None
        self.update_annotation_fields()
        self.update_plot()

    def delete_annotation(self):
        if self.state.active_event is None:
            return
        self.signals.delete_event(self.state.active_event)
        self.state.active_event = None
        self.update_annotation_fields()
        self.update_plot()

    def build_tk_ui(self):
        self.root = tk.Tk()
        self.root.title('脑电信号可视化')
        # self.root.geometry('1000x600')

        # ========================= 信息显示区  ========================= #
        info_frame = ttk.Frame(self.root)
        info_frame.pack(side=tk.TOP, fill=tk.X)

        ttk.Label(info_frame, text='信号名称: ' + self.signals.name).pack(side=tk.LEFT)
        ttk.Label(info_frame, text='通道标签: ' + ', '.join(self.signals.channels)).pack(
            side=tk.LEFT, padx=5
        )
        ttk.Label(info_frame, text='采样率: ' + str(self.signals.srate)).pack(
            side=tk.LEFT, padx=5
        )
        ttk.Label(info_frame, text='时间长度: ' + str(self.signals.t_duration)).pack(
            side=tk.LEFT, padx=5
        )

        # 退出按钮
        ttk.Button(info_frame, text='退出', command=self.root.quit).pack(side=tk.RIGHT)
        ttk.Button(info_frame, text='保存', command=self.save_annotations).pack(
            side=tk.RIGHT
        )

        # ========================= 图形区  ========================= #
        plot_frame = ttk.Frame(self.root)
        plot_frame.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        fig = plt.figure()
        self.canvas = FigureCanvasTkAgg(fig, master=plot_frame)
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        # ========================= 按钮区  ========================= #
        button_frame = ttk.Frame(self.root)
        button_frame.pack(side=tk.BOTTOM, fill=tk.X)

        # --------------- 下部左侧 --------------- #
        ttk.Button(button_frame, text='向左加载', command=self.load_left).pack(
            side=tk.LEFT
        )
        ttk.Button(button_frame, text='向右加载', command=self.load_right).pack(
            side=tk.LEFT
        )

        def update_state(onset=None, duration=None, interval=None):
            self.state.t_start = onset or self.state.t_start
            self.state.t_page = duration or self.state.t_page
            self.state.interval = interval or self.state.interval

        ttk.Label(button_frame, text='开始时间:').pack(side=tk.LEFT)
        self.onset_entry = ttk.Entry(button_frame)
        self.onset_entry.pack(side=tk.LEFT)
        self.onset_entry.bind(
            '<KeyRelease>',
            lambda event: update_state(onset=float(self.onset_entry.get())),
        )

        ttk.Label(button_frame, text='持续时间:').pack(side=tk.LEFT)
        self.duration_entry = ttk.Entry(button_frame)
        self.duration_entry.pack(side=tk.LEFT)
        self.duration_entry.bind(
            '<KeyRelease>',
            lambda event: update_state(duration=float(self.duration_entry.get())),
        )

        ttk.Label(button_frame, text='Y轴 Interval:').pack(side=tk.LEFT)
        self.interval_entry = ttk.Entry(button_frame)
        self.interval_entry.pack(side=tk.LEFT)

        def _get_interval():
            interval = self.interval_entry.get()
            if interval:
                return float(interval)
            return 'auto'  # interval 可以为 auto

        self.interval_entry.bind(
            '<KeyRelease>', lambda event: update_state(interval=_get_interval())
        )
        ttk.Button(button_frame, text='刷新', command=self.update_plot).pack(side=tk.LEFT)

        ttk.Label(button_frame, text='\t').pack(side=tk.LEFT)
        # --------------- 下部右侧 --------------- #

        ttk.Label(button_frame, text='标注标签:').pack(side=tk.LEFT)
        self.annt_label_entry = ttk.Entry(button_frame)
        self.annt_label_entry.pack(side=tk.LEFT)

        ttk.Label(button_frame, text='标注开始时间:').pack(side=tk.LEFT)
        self.annt_onset_entry = ttk.Entry(button_frame)
        self.annt_onset_entry.pack(side=tk.LEFT)

        ttk.Label(button_frame, text='标注持续时间:').pack(side=tk.LEFT)
        self.annt_duration_entry = ttk.Entry(button_frame)
        self.annt_duration_entry.pack(side=tk.LEFT)

        ttk.Button(button_frame, text='添加标注', command=self.add_annotation).pack(
            side=tk.LEFT
        )
        ttk.Button(button_frame, text='删除标注', command=self.delete_annotation).pack(
            side=tk.LEFT
        )

        # 绑定图形点击事件
        fig.canvas.mpl_connect('button_press_event', self.on_click_figure)

    def run(self):
        self.build_tk_ui()
        self.update_entry_fields()
        self.update_plot()  # 初始化时更新绘图
        self.root.mainloop()
        # self.fn_save()  # 保存功能的回调

    def save_annotations(self):
        event_fpath = self.signals.filepath.with_suffix('.events.json')

        write_json(event_fpath, [event.to_dict() for event in self.signals.events])
        print(f'保存标注信息到: {event_fpath}')


def _launch_app(signals: EEGFileInterface, fn_on_exit: Callable[[], None] | None = None):
    app = SignalAnnotatorApp(signals, fn_on_exit)
    app.run()


def launch_app(fpath: str | None = None, srate: float | None = None):
    from zyplib.io import EDFFileInterface, MatFileInterface

    from .select_file import launch_app as launch_select_file_app

    def on_file_selected(path: Path):
        print_debug(f'Viz 文件: {path}')
        if path.suffix == '.mat':
            signals = MatFileInterface(path, srate=srate)
        elif path.suffix == '.edf':
            signals = EDFFileInterface(path)
        else:
            raise ValueError(f'不支持的文件类型: {path}')

        _launch_app(signals)

    if fpath is None:
        launch_select_file_app(on_file_selected)
    else:
        on_file_selected(Path(fpath))



from torchmetrics import Metric, classification, regression

from .torch_metric import TorchMetricRecorder


def clf_metrics_builder(binary: bool = True) -> dict[str, Metric]:
    """构建并返回分类指标字典。

    Parameters:
    ----------
    - `binary` : `bool`, 可选
        - 指示分类任务是否为二分类。如果为 True，则使用二分类指标。
          如果为 False，则使用一般分类指标。默认为 True。

    Return:
    ----------
    - `dict[str, Metric]`, 键是指标名称（例如 "accuracy", "precision"），值是相应的 `Metric` 对象，
    """
    if binary:
        metrics = {
            'accuracy': classification.BinaryAccuracy(),
            'precision': classification.BinaryPrecision(),
            'recall': classification.BinaryRecall(),
            'f1_score': classification.BinaryF1Score(),
        }
    else:
        metrics = {
            'accuracy': classification.Accuracy(),
            'precision': classification.Precision(),
            'recall': classification.Recall(),
            'f1_score': classification.F1Score(),
        }

    return metrics


def reg_metrics_builder() -> dict[str, Metric]:
    """构建并返回回归指标字典。

    Return:
    ----------
    - `dict[str, Metric]`, 键是指标名称（例如 "mae", "mse"），值是相应的 `Metric` 对象。
    """
    metrics = {
        'mae': regression.MeanAbsoluteError(),
        'mse': regression.MeanSquaredError(),
        'rmse': regression.MeanSquaredError(squared=False),  # RMSE 是 MSE 的平方根
        'r2_score': regression.R2Score(),
    }
    return metrics


__all__ = ['TorchMetricRecorder', 'clf_metrics_builder', 'reg_metrics_builder']



from collections import Counter

import numpy as np

from zyplib.utils.ensure import ensure_npy
from zyplib.utils.print import print_info


def summary_labels(y: np.ndarray, just_print: bool = False) -> Counter:
    y = ensure_npy(y)
    y = y.squeeze()
    if y.ndim != 1:
        raise ValueError('y.ndim must be 1')

    counter = Counter(y)

    if not just_print:
        return counter

    print_info(counter)
